# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from typing import Iterable

import numpy as np

from wulfric.cell._basic_manipulation import get_reciprocal

from wulfric.constants._hpkot_convention import HPKOT_DEFAULT_K_PATHS
from wulfric.constants._sc_convention import SC_DEFAULT_K_PATHS
from wulfric.constants._space_groups import INVERSION_SYMMETRY

from wulfric.crystal._crystal_validation import validate_atoms
from wulfric.crystal._conventional import get_conventional
from wulfric.crystal._primitive import get_primitive
from wulfric.crystal._sc_variation import sc_get_variation
from wulfric.crystal._hpkot_extended_bl_symbol import hpkot_get_extended_bl_symbol

from wulfric._exceptions import ConventionNotSupported

from wulfric.kpoints._hpkot_points import _hpkot_get_points
from wulfric.kpoints._sc_points import _sc_get_points

from wulfric._spglib_interface import get_spglib_data, validate_spglib_data
from wulfric._syntactic_sugar import SyntacticSugar


# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_path_as_list(path_as_string) -> list:
    r"""
    Converts k=path from string to list representation.

    .. code-block:: python

        path_as_list = [["G", "X", "Y", "S"], ["G", "Z"]]
        path_as_string = "G-X-Y-S|G-Z"

    Parameters
    ----------
    path_as_string : str
        K-path as a single string.

    Returns
    -------
    path_as_list : list of list of str
        K-path as list of subpaths, where each subpath is a list of k-points.

    Raises
    ------
    ValueError
        If ``path_as_string`` is not ``str``
    Valueerror
        If any subpath contains less than two points.

    See Also
    --------
    :ref:`user-guide_usage_key-concepts_k-path`
    """
    if not isinstance(path_as_string, str):
        raise ValueError(f"path_as_string is not a string: {path_as_string}")

    subpaths_as_str = path_as_string.split("|")
    path_as_list = []
    for s_i, subpath_as_str in enumerate(subpaths_as_str):
        subpath_as_list = subpath_as_str.split("-")
        # Each subpath has to contain at least two points.
        # If subpath is empty, then no action required
        if len(subpath_as_list) == 1:
            raise ValueError(
                f"Subpath {s_i + 1} has less than two points: {subpath_as_str}"
            )

        if len(subpath_as_list) >= 2:
            path_as_list.append(subpath_as_list)

    return path_as_list


def get_path_as_string(path_as_list) -> str:
    r"""
    Converts k=path from list to string representation.

    .. code-block:: python

        path_as_string = "G-X-Y-S|G-Z"
        path_as_list = [["G", "X", "Y", "S"], ["G", "Z"]]

    Parameters
    ----------
    path_as_list : list of list of str
        K-path as list of subpaths, where each subpath is a list of k-points.


    Returns
    -------
    path_as_string : str
        K-path as a single string.

    Raises
    ------
    ValueError
        If ``path_as_list`` is not a ``list`` of ``list``.

    See Also
    --------
    :ref:`user-guide_usage_key-concepts_k-path`
    """

    if not isinstance(path_as_list, Iterable):
        raise ValueError(
            f"path_as_list is not Iterable, can not convert to string:\n{path_as_list}"
        )

    subpaths_as_str = []
    for s_i, subpath_as_list in enumerate(path_as_list):
        if not isinstance(subpath_as_list, Iterable):
            raise ValueError(
                f"Subpath {s_i + 1} is not Iterable, can not convert to string:\n{subpath_as_list}"
            )
        subpaths_as_str.append("-".join(subpath_as_list))

    return "|".join(subpaths_as_str)


def get_path_and_points(
    cell,
    atoms,
    spglib_data=None,
    convention="HPKOT",
    with_time_reversal=True,
    relative=True,
):
    r"""
    Returns recommended k-path and set of high-symmetry points.

    Note that high-symmetry points are the ones of the primitive cell, that is associated
    with the given set of ``cell`` and ``atoms``. In other words it respects the symmetry
    of the crystal.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_

            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional

            See Notes
        *   "species" : (N, ) list of str, optional

            See Notes
        *   "spglib_types" : (N, ) list of int, optional

            See Notes

        .. hint::
            Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
            like to interpret the ``cell`` alone (effectively assuming that the ``cell``
            is a primitive one).

    spglib_data : :py:class:`.SyntacticSugar`, optional
        If you need more control on the parameters passed to the spglib, then
        you can get ``spglib_data`` manually and pass it to this function.
        Use wulfric's interface to |spglib|_ as

        .. code-block:: python

            spglib_data = wulfric.get_spglib_data(...)

        using the same ``cell`` and ``atoms["positions"]`` that you are passing to this
        function.

    convention : str, default "HPKOT"
        Convention for the definition of the conventional cell. Case-insensitive.
        Supported:

        * "HPKOT" for [1]_
        * "SC" for [2]_

    with_time_reversal : bool, default True
        Whether to assume that the system has time reversal symmetry. By default
        assumes that the system has it. The strategy for extending the path when
        ``with_time_reversal=False`` for the crystals without inversion symmetry is
        described in [1]_. For the systems with inversion symmetry this parameter does
        nothing.

        .. versionadded:: 0.6.3

    relative : bool, default True
        Whether to return coordinates as relative to the reciprocal cell or in absolute
        coordinates in the reciprocal Cartesian space.

    Returns
    -------
    recommended_path : str
        Recommended path in reciprocal space between the high-symmetry k points.
    hs_points : dict
        High symmetry points.

        .. code-block:: python

            hs_points = {name1: coordinate1, name2 : coordinate2, ...}

        Coordinates of the points are

        * (default) Relative to ``get_reciprocal(cell)`` if ``relative=True``.
        * Absolute in reciprocal space if ``relative=False``.

    Notes
    =====
    |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces the
    ``types`` for given atoms see :py:func:`wulfric.get_spglib_types`.


    References
    ==========
    .. [1] Hinuma, Y., Pizzi, G., Kumagai, Y., Oba, F. and Tanaka, I., 2017.
           Band structure diagram paths based on crystallography.
           Computational Materials Science, 128, pp.140-184.
    .. [2] Setyawan, W. and Curtarolo, S., 2010.
           High-throughput electronic band structure calculations: Challenges and tools.
           Computational materials science, 49(2), pp. 299-312.

    See Also
    --------
    wulfric.Kpoints : Class with a convenient interface for the same information.

    """

    cell = np.array(cell, dtype=float)

    # Validate that the atoms dictionary is what expected of it
    validate_atoms(atoms=atoms, required_keys=["positions"], raise_errors=True)

    # Call spglib
    if spglib_data is None:
        spglib_data = get_spglib_data(cell=cell, atoms=atoms)
    # Or check that spglib_data were *most likely* produced via wulfric's interface
    elif not isinstance(spglib_data, SyntacticSugar):
        raise TypeError(
            f"Are you sure that spglib_data were produced via wulfric's interface? Expected SyntacticSugar, got {type(spglib_data)}."
        )
    # Validate that user-provided spglib_data match user-provided structure
    else:
        validate_spglib_data(cell=cell, atoms=atoms, spglib_data=spglib_data)

    conventional_cell, _ = get_conventional(
        cell=cell, atoms=atoms, convention=convention, spglib_data=spglib_data
    )
    primitive_cell, _ = get_primitive(
        cell=cell, atoms=atoms, convention=convention, spglib_data=spglib_data
    )

    lattice_type = spglib_data.crystal_family + spglib_data.centring_type

    convention = convention.lower()
    if convention == "sc":
        lattice_variation = sc_get_variation(
            cell=cell, atoms=atoms, spglib_data=spglib_data
        )
        hs_points = _sc_get_points(
            conventional_cell=conventional_cell,
            lattice_type=lattice_type,
            lattice_variation=lattice_variation,
        )
        kpath = SC_DEFAULT_K_PATHS[lattice_variation]
    elif convention == "hpkot":
        extended_bl_symbol = hpkot_get_extended_bl_symbol(
            cell=cell, atoms=atoms, spglib_data=spglib_data
        )
        hs_points = _hpkot_get_points(
            conventional_cell=conventional_cell,
            lattice_type=lattice_type,
            extended_bl_symbol=extended_bl_symbol,
        )
        kpath = HPKOT_DEFAULT_K_PATHS[extended_bl_symbol]
    else:
        raise ConventionNotSupported(convention, supported_conventions=["HPKOT", "SC"])

    primitive_rcell = get_reciprocal(cell=primitive_cell)

    for point in hs_points:
        # Here coordinates are absolute
        hs_points[point] = hs_points[point] @ primitive_rcell

        if relative:
            # absolute -> relative
            hs_points[point] = hs_points[point] @ np.linalg.inv(
                get_reciprocal(cell=cell)
            )

    if (
        not INVERSION_SYMMETRY[spglib_data.space_group_number]
        and not with_time_reversal
    ):
        # Extend k-path
        extra_path = get_path_as_list(path_as_string=kpath)
        for i in range(len(extra_path)):
            for j in range(len(extra_path[i])):
                if extra_path[i][j] != "GAMMA":
                    extra_path[i][j] = f"{extra_path[i][j]}_P"
        kpath = f"{kpath}|{get_path_as_string(path_as_list=extra_path)}"

        # Extend high-symmetry points
        points = list(hs_points)
        for point in points:
            if point != "GAMMA":
                hs_points[f"{point}_P"] = -hs_points[point]

    return kpath, hs_points


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from math import cos, sin, sqrt

import numpy as np

from wulfric.cell._basic_manipulation import get_params, get_reciprocal
from wulfric.crystal._crystal_validation import validate_atoms
from wulfric._spglib_interface import get_spglib_data, validate_spglib_data
from wulfric._syntactic_sugar import SyntacticSugar
from wulfric.crystal._conventional import get_conventional
from wulfric._exceptions import PotentialBugError
from wulfric.constants import TORADIANS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def hpkot_get_extended_bl_symbol(cell, atoms, spglib_data=None):
    r"""
    Returns extended bravais lattice symbol as defined in the paper by Hinuma, Pizzi,
    Kumagai, Oba, and Tanaka [1]_.

    .. versionadded:: 0.6.3

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    atoms : dict
        Dictionary with N atoms. Expected keys:

        *   "positions" : (N, 3) |array-like|_

            Positions of the atoms in the basis of lattice vectors (``cell``). In other
            words - relative coordinates of atoms.
        *   "names" : (N, ) list of str, optional

            See Notes
        *   "species" : (N, ) list of str, optional

            See Notes
        *   "spglib_types" : (N, ) list of int, optional

            See Notes

        .. hint::
            Pass ``atoms = dict(positions=[[0, 0, 0]], spglib_types=[1])`` if you would
            like to interpret the ``cell`` alone (effectively assuming that the ``cell``
            is a primitive one).

    spglib_data : :py:class:`.SyntacticSugar`, optional
        If you need more control on the parameters passed to the spglib, then
        you can get ``spglib_data`` manually and pass it to this function.
        Use wulfric's interface to |spglib|_ as

        .. code-block:: python

            spglib_data = wulfric.get_spglib_data(...)

        using the same ``cell`` and ``atoms["positions"]`` that you are passing to this
        function.

    Returns
    -------
    extended_bl_type : str
        Extended Bravais lattice symbol.

    Notes
    -----
    |spglib|_ uses ``types`` to distinguish the atoms. To see how wulfric deduces the
    ``types`` for given atoms see :py:func:`wulfric.get_spglib_types`.

    References
    ----------
    .. [1] Hinuma, Y., Pizzi, G., Kumagai, Y., Oba, F. and Tanaka, I., 2017.
           Band structure diagram paths based on crystallography.
           Computational Materials Science, 128, pp.140-184.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.crystal.hpkot_get_extended_bl_symbol(
        ...     cell=[[1, 0, 0], [0, 1, 0], [0, 0, 1]],
        ...     atoms=dict(positions=[[0, 0, 0]], spglib_types=[1]),
        ... )
        'cP2'

    """

    # Validate that the atoms dictionary is what expected of it
    validate_atoms(atoms=atoms, required_keys=["positions"], raise_errors=True)

    # Call spglib
    if spglib_data is None:
        spglib_data = get_spglib_data(cell=cell, atoms=atoms)
    # Or check that spglib_data were *most likely* produced via wulfric's interface
    elif not isinstance(spglib_data, SyntacticSugar):
        raise TypeError(
            f"Are you sure that spglib_data were produced via wulfric's interface? Expected SyntacticSugar, got {type(spglib_data)}."
        )
    # Validate that user-provided spglib_data match user-provided structure
    else:
        validate_spglib_data(cell=cell, atoms=atoms, spglib_data=spglib_data)

    cell = np.array(cell, dtype=float)

    lattice_type = spglib_data.crystal_family + spglib_data.centring_type

    # Lattice types that do not require computation of lattice parameters
    if lattice_type in ["cI", "tP", "oP", "mP"]:
        return f"{lattice_type}1"

    if lattice_type == "cP":
        if 195 <= spglib_data.space_group_number <= 206:
            return "cP1"
        elif 207 <= spglib_data.space_group_number <= 230:
            return "cP2"
        else:
            raise PotentialBugError(
                error_summary=f'(convention="HPKOT"), lattice type cP, space group {spglib_data.space_group_number}. Failed to define extended Bravais lattice symbol.'
            )

    if lattice_type == "cF":
        if 195 <= spglib_data.space_group_number <= 206:
            return "cF1"
        elif 207 <= spglib_data.space_group_number <= 230:
            return "cF2"
        else:
            raise PotentialBugError(
                error_summary=f'(convention="HPKOT"), lattice type cF, space group {spglib_data.space_group_number}. Failed to define extended Bravais lattice symbol.'
            )

    if lattice_type == "hP":
        if (
            143 <= spglib_data.space_group_number <= 149
            or 159 <= spglib_data.space_group_number <= 163
            or spglib_data.space_group_number in [151, 153, 157]
        ):
            return "hP1"
        else:
            return "hP2"

    # Lattice types that require computation of lattice parameters
    conventional_cell, _ = get_conventional(
        cell=cell, atoms=atoms, convention="HPKOT", spglib_data=spglib_data
    )
    a, b, c, _, beta, _ = get_params(cell=conventional_cell)

    if lattice_type == "tI":
        if c <= a:
            return "tI1"
        else:
            return "tI2"

    if lattice_type == "oF":
        if 1 / a**2 > 1 / b**2 + 1 / c**2:
            return "oF1"
        elif 1 / c**2 > 1 / a**2 + 1 / b**2:
            return "oF2"
        else:
            return "oF3"

    if lattice_type == "oI":
        if c >= a and c >= b:
            return "oI1"
        if a >= b and a >= c:
            return "oI2"
        if b >= a and b >= c:
            return "oI3"

    if lattice_type == "oC":
        if a <= b:
            return "oC1"
        else:
            return "oC2"

    if lattice_type == "oA":
        if b <= c:
            return "oA1"
        else:
            return "oA2"

    if lattice_type == "hR":
        if sqrt(3) * a <= sqrt(2) * c:
            return "hR1"
        else:
            return "hR2"

    if lattice_type == "mC":
        if b <= a * sin(beta * TORADIANS):
            return "mC1"
        else:
            if (
                -a * cos(beta * TORADIANS) / c + ((a * sin(beta * TORADIANS)) / b) ** 2
                <= 1
            ):
                return "mC2"
            else:
                return "mC3"

    if lattice_type == "aP":
        _, _, _, r_alpha, r_beta, r_gamma = get_params(
            cell=get_reciprocal(cell=conventional_cell)
        )

        if r_alpha >= 90 and r_beta >= 90 and r_gamma >= 90:
            return "aP2"
        else:
            return "aP3"

    # If lattice type is not one of the expected ones
    raise PotentialBugError(
        f'(convention="HPKOT"), lattice type {lattice_type}, space group {spglib_data.space_group_number}.. Failed to identify lattice type (not one of supported).'
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from math import cos
from math import pi as PI
from math import sin, sqrt

import numpy as np

from wulfric.cell._basic_manipulation import from_params, get_reciprocal
from wulfric.constants._numerical import TORADIANS
from wulfric.constants._sc_convention import SC_BRAVAIS_LATTICE_VARIATIONS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


# Primitive cell`s construction
def SC_CUB(a: float):
    r"""
    Constructs primitive cubic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a, &0, &0)\\
            \boldsymbol{a}_2 &=& (0, &a, &0)\\
            \boldsymbol{a}_3 &=& (0, &0, &a)
        \end{matrix}

    Parameters
    ----------
    a : float or int
        Length of the three lattice vectors of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_CUB(a=2)
        array([[2, 0, 0],
               [0, 2, 0],
               [0, 0, 2]])
    """

    return np.array([[a, 0, 0], [0, a, 0], [0, 0, a]])


def SC_FCC(a: float):
    r"""
    Constructs primitive face-centred cubic cell as defined in [1]_.

    .. math::

        \begin{matrix}
        \boldsymbol{a}_1 &=& (0, &a/2, &a/2)\\
        \boldsymbol{a}_2 &=& (a/2, &0, &a/2)\\
        \boldsymbol{a}_3 &=& (a/2, &a/2, &0)
        \end{matrix}

    Parameters
    ----------
    a : float
        Length of the three lattice vectors of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_FCC(a=2)
        array([[0., 1., 1.],
               [1., 0., 1.],
               [1., 1., 0.]])
    """

    return np.array([[0, a / 2, a / 2], [a / 2, 0, a / 2], [a / 2, a / 2, 0]])


def SC_BCC(a: float):
    r"""
    Constructs primitive body-centred cubic cell as defined in [1]_.

    .. math::

        \begin{matrix}
        \boldsymbol{a}_1 &=& (-a/2,& a/2,& a/2)\\
        \boldsymbol{a}_2 &=& (a/2, &-a/2,& a/2)\\
        \boldsymbol{a}_3 &=& (a/2, &a/2, &-a/2)
        \end{matrix}

    Parameters
    ----------
    a : float
        Length of the three lattice vectors of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_BCC(a=2)
        array([[-1.,  1.,  1.],
               [ 1., -1.,  1.],
               [ 1.,  1., -1.]])
    """

    return np.array(
        [[-a / 2, a / 2, a / 2], [a / 2, -a / 2, a / 2], [a / 2, a / 2, -a / 2]]
    )


def SC_TET(a: float, c: float):
    r"""
    Constructs primitive tetragonal cell as defined in [1]_.

    .. math::

        \begin{matrix}
        \boldsymbol{a}_1 &=& (a, &0, &0)\\
        \boldsymbol{a}_2 &=& (0, &a, &0)\\
        \boldsymbol{a}_3 &=& (0, &0, &c)
        \end{matrix}

    Parameters
    ----------
    a : float
        Length of the first two lattice vectors of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_TET(a=2, c=5)
        array([[2, 0, 0],
               [0, 2, 0],
               [0, 0, 5]])
    """

    return np.array([[a, 0, 0], [0, a, 0], [0, 0, c]])


def SC_BCT(a: float, c: float):
    r"""
    Constructs primitive body-centred tetragonal cell as defined in [1]_.

    .. math::

        \begin{matrix}
        \boldsymbol{a}_1 &=& (-a/2, &a/2, &c/2)\\
        \boldsymbol{a}_2 &=& (a/2, &-a/2, &c/2)\\
        \boldsymbol{a}_3 &=& (a/2, &a/2, &-c/2)
        \end{matrix}

    Parameters
    ----------
    a : float
        Length of the first two lattice vectors of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_BCT(a=2, c=5)
        array([[-1. ,  1. ,  2.5],
               [ 1. , -1. ,  2.5],
               [ 1. ,  1. , -2.5]])
    """

    return np.array(
        [[-a / 2, a / 2, c / 2], [a / 2, -a / 2, c / 2], [a / 2, a / 2, -c / 2]]
    )


def SC_ORC(a: float, b: float, c: float):
    r"""
    Constructs primitive orthorhombic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a, &0, &0)\\
            \boldsymbol{a}_2 &=& (0, &b, &0)\\
            \boldsymbol{a}_3 &=& (0, &0, &c)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_ORC(a=3, b=5, c=7)
        array([[3, 0, 0],
               [0, 5, 0],
               [0, 0, 7]])
    """

    return np.array([[a, 0, 0], [0, b, 0], [0, 0, c]])


def SC_ORCF(a: float, b: float, c: float):
    r"""
    Constructs primitive face-centred orthorhombic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (0, &b/2, &c/2)\\
            \boldsymbol{a}_2 &=& (a/2, &0, &c/2)\\
            \boldsymbol{a}_3 &=& (a/2, &b/2, &0)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_ORCF(a=3, b=5, c=7)
        array([[0. , 2.5, 3.5],
               [1.5, 0. , 3.5],
               [1.5, 2.5, 0. ]])
    """

    return np.array([[0, b / 2, c / 2], [a / 2, 0, c / 2], [a / 2, b / 2, 0]])


def SC_ORCI(a: float, b: float, c: float):
    r"""
    Constructs primitive body-centred orthorhombic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (-a/2, &b/2, &c/2)\\
            \boldsymbol{a}_2 &=& (a/2, &-b/2, &c/2)\\
            \boldsymbol{a}_3 &=& (a/2, &b/2, &-c/2)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_ORCI(a=3, b=5, c=7)
        array([[-1.5,  2.5,  3.5],
               [ 1.5, -2.5,  3.5],
               [ 1.5,  2.5, -3.5]])
    """

    return np.array(
        [[-a / 2, b / 2, c / 2], [a / 2, -b / 2, c / 2], [a / 2, b / 2, -c / 2]]
    )


def SC_ORCC(a: float, b: float, c: float):
    r"""
    Constructs primitive base-centred orthorhombic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a/2, &-b/2, &0)\\
            \boldsymbol{a}_2 &=& (a/2, &b/2, &0)\\
            \boldsymbol{a}_3 &=& (0, &0, &c)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_ORCC(a=3, b=5, c=7)
        array([[ 1.5, -2.5,  0. ],
               [ 1.5,  2.5,  0. ],
               [ 0. ,  0. ,  7. ]])
    """

    return np.array([[a / 2, -b / 2, 0], [a / 2, b / 2, 0], [0, 0, c]])


def SC_HEX(a: float, c: float):
    r"""
    Constructs primitive hexagonal cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (\frac{a}{2}, &\frac{-a\sqrt{3}}{2}, &0)\\
            \boldsymbol{a}_2 &=& (\frac{a}{2}, &\frac{a\sqrt{3}}{2}, &0)\\
            \boldsymbol{a}_3 &=& (0, &0, &c)
        \end{matrix}

    Parameters
    ----------
    a : float
        Length of the first two lattice vectors of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_HEX(a=3, c=5)
        array([[ 1.5       , -2.59807621,  0.        ],
               [ 1.5       ,  2.59807621,  0.        ],
               [ 0.        ,  0.        ,  5.        ]])
    """

    return np.array(
        [[a / 2, -a * sqrt(3) / 2, 0], [a / 2, a * sqrt(3) / 2, 0], [0, 0, c]]
    )


def SC_RHL(a: float, alpha: float):
    r"""
    Constructs primitive rhombohedral cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a\cos(\alpha / 2), &-a\sin(\alpha/2), &0)\\
            \boldsymbol{a}_2 &=& (a\cos(\alpha / 2), &a\sin(\alpha/2), &0)\\
            \boldsymbol{a}_3 &=& \left(\dfrac{\cos\alpha}{\cos(\alpha/2)}\right.,
            &0, &\left.a\sqrt{1 - \dfrac{\cos^2\alpha}{\cos^2(\alpha/2)}}\right)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the lattice vectors of the conventional cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_RHL(a=3, alpha=40)
        array([[ 2.81907786, -1.02606043,  0.        ],
               [ 2.81907786,  1.02606043,  0.        ],
               [ 2.44562241,  0.        ,  1.73750713]])
    """

    alpha *= TORADIANS
    return np.array(
        [
            [a * cos(alpha / 2), -a * sin(alpha / 2), 0],
            [a * cos(alpha / 2), a * sin(alpha / 2), 0],
            [
                a * cos(alpha) / cos(alpha / 2),
                0,
                a * sqrt(1 - cos(alpha) ** 2 / cos(alpha / 2) ** 2),
            ],
        ]
    )


def SC_MCL(a: float, b: float, c: float, alpha: float):
    r"""
    Constructs primitive monoclinic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a, &0, &0)\\
            \boldsymbol{a}_2 &=& (0, &b, &0)\\
            \boldsymbol{a}_3 &=& (0, &c\cos\alpha, &c\sin\alpha)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second of the two remaining lattice vectors of the conventional
        cell.
    c : float
        Length of the third of the two remaining lattice vectors of the conventional cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_MCL(a=3, b=5, c=7, alpha=45)
        array([[3.        , 0.        , 0.        ],
               [0.        , 5.        , 0.        ],
               [0.        , 4.94974747, 4.94974747]])
    """

    alpha *= TORADIANS
    return np.array([[a, 0, 0], [0, b, 0], [0, c * cos(alpha), c * sin(alpha)]])


def SC_MCLC(a: float, b: float, c: float, alpha: float):
    r"""
    Constructs primitive base-centred monoclinic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a/2, &b/2, &0)\\
            \boldsymbol{a}_2 &=& (-a/2, &b/2, &0)\\
            \boldsymbol{a}_3 &=& (0, &c\cos\alpha, &c\sin\alpha)
        \end{matrix}

    Input values are used as they are, therefore, the cell might not be a standard
    primitive one.

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second of the two remaining lattice vectors of the conventional
        cell.
    c : float
        Length of the third of the two remaining lattice vectors of the conventional
        cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_MCLC(a=3, b=5, c=7, alpha=45)
        array([[ 1.5       ,  2.5       ,  0.        ],
               [-1.5       ,  2.5       ,  0.        ],
               [ 0.        ,  4.94974747,  4.94974747]])
    """

    alpha *= TORADIANS
    return np.array(
        [
            [a / 2, b / 2, 0],
            [-a / 2, b / 2, 0],
            [0, c * cos(alpha), c * sin(alpha)],
        ]
    )


def SC_TRI(
    a: float,
    b: float,
    c: float,
    alpha: float,
    beta: float,
    gamma: float,
    input_reciprocal=False,
):
    r"""
    Constructs primitive triclinic cell as defined in [1]_.

    .. math::

        \begin{matrix}
            \boldsymbol{a}_1 &=& (a, &0, &0)\\
            \boldsymbol{a}_2 &=& (b\cos\gamma, &b\sin\gamma, &0)\\
            \boldsymbol{a}_3 &=& (c\cos\beta, &\dfrac{c(\cos\alpha - \cos\beta\cos\gamma)}{\sin\gamma}, &\dfrac{c}{\sin\gamma}\sqrt{\sin^2\gamma - \cos^2\alpha - \cos^2\beta + 2\cos\alpha\cos\beta\cos\gamma})
        \end{matrix}

    Parameters
    ----------
    a : float
        Length of the first lattice vector of the conventional cell.
    b : float
        Length of the second lattice vector of the conventional cell.
    c : float
        Length of the third lattice vector of the conventional cell.
    alpha : float
        Angle between vectors :math:`a_2` and :math:`a_3` of the conventional cell in
        degrees.
    beta : float
        Angle between vectors :math:`a_1` and :math:`a_3` of the conventional cell in
        degrees.
    gamma : float
        Angle between vectors :math:`a_1` and :math:`a_2` of the conventional cell in
        degrees.
    input_reciprocal : bool, default False
        Whether to interpret input as reciprocal parameters.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a primitive cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.SC_TRI(a=3, b=5, c=7, alpha=45, beta=33, gamma=21)
        array([[ 3.        ,  0.        ,  0.        ],
               [ 4.66790213,  1.79183975,  0.        ],
               [ 5.87069398, -1.48176621,  3.51273699]])
    """

    cell = from_params(a, b, c, alpha, beta, gamma)
    if input_reciprocal:
        cell = get_reciprocal(cell)

    return cell


def sc_get_example(lattice_variation: str = None):
    r"""
    Examples of the Bravais lattices as defined in the paper by Setyawan and Curtarolo [1]_.

    .. versionchanged:: 0.6.3 renamed from ``sc_get_example_cell``.

    Parameters
    ----------
    lattice_variation : str, optional
        Name of the lattice type or variation to be returned. For available names see
        documentation of each :ref:`user-guide_conventions_bravais-lattices`.
        Case-insensitive.

    Returns
    -------
    cell : (3, 3) :numpy:`ndarray`
        Matrix of a direct cell, rows are interpreted as vectors.

        .. code-block:: python

            cell = [
                [a1_x, a1_y, a1_z],
                [a2_x, a2_y, a2_z],
                [a3_x, a3_y, a3_z],
            ]

    References
    ----------
    .. [1] Setyawan, W. and Curtarolo, S., 2010.
        High-throughput electronic band structure calculations: Challenges and tools.
        Computational materials science, 49(2), pp. 299-312.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.cell.sc_get_example("cub")
        array([[3.14159265, 0.        , 0.        ],
               [0.        , 3.14159265, 0.        ],
               [0.        , 0.        , 3.14159265]])
        >>> wulfric.cell.sc_get_example("ORCF3")
        array([[0.        , 1.96349541, 2.61799388],
               [1.57079633, 0.        , 2.61799388],
               [1.57079633, 1.96349541, 0.        ]])
    """

    correct_inputs = set(map(lambda x: x.lower(), SC_BRAVAIS_LATTICE_VARIATIONS)).union(
        set(
            map(
                lambda x: x.translate(str.maketrans("", "", "12345ab")).lower(),
                SC_BRAVAIS_LATTICE_VARIATIONS,
            )
        )
    )

    if (
        not isinstance(lattice_variation, str)
        or lattice_variation.lower() not in correct_inputs
    ):
        message = (
            f'There is no example of "{lattice_variation}" Bravais lattice. '
            "Available examples are:\n"
        )
        for name in correct_inputs:
            message += f"  * {name}\n"
        raise ValueError(message)

    lattice_variation = lattice_variation.lower()

    if lattice_variation == "cub":
        cell = SC_CUB(PI)
    elif lattice_variation == "fcc":
        cell = SC_FCC(PI)
    elif lattice_variation == "bcc":
        cell = SC_BCC(PI)
    elif lattice_variation == "tet":
        cell = SC_TET(PI, 1.5 * PI)
    elif lattice_variation in ["bct1", "bct"]:
        cell = SC_BCT(1.5 * PI, PI)
    elif lattice_variation == "bct2":
        cell = SC_BCT(PI, 1.5 * PI)
    elif lattice_variation == "orc":
        cell = SC_ORC(PI, 1.5 * PI, 2 * PI)
    elif lattice_variation in ["orcf1", "orcf"]:
        cell = SC_ORCF(0.7 * PI, 5 / 4 * PI, 5 / 3 * PI)
    elif lattice_variation == "orcf2":
        cell = SC_ORCF(1.2 * PI, 5 / 4 * PI, 5 / 3 * PI)
    elif lattice_variation == "orcf3":
        cell = SC_ORCF(PI, 5 / 4 * PI, 5 / 3 * PI)
    elif lattice_variation == "orci":
        return SC_ORCI(PI, 1.3 * PI, 1.7 * PI)
    elif lattice_variation == "orcc":
        cell = SC_ORCC(PI, 1.3 * PI, 1.7 * PI)
    elif lattice_variation == "hex":
        cell = SC_HEX(PI, 2 * PI)
    elif lattice_variation in ["rhl1", "rhl"]:
        # If alpha = 60 it is effectively FCC!
        cell = SC_RHL(PI, 70)
    elif lattice_variation == "rhl2":
        cell = SC_RHL(PI, 110)
    elif lattice_variation == "mcl":
        cell = SC_MCL(PI, 1.3 * PI, 1.6 * PI, alpha=75)
    elif lattice_variation in ["mclc1", "mclc"]:
        cell = SC_MCLC(PI, 1.4 * PI, 1.7 * PI, 80)
    elif lattice_variation == "mclc2":
        cell = SC_MCLC(1.4 * PI * sin(75 * TORADIANS), 1.4 * PI, 1.7 * PI, 75)
    elif lattice_variation == "mclc3":
        b = PI
        x = 1.1
        alpha = 78
        ralpha = alpha * TORADIANS
        c = b * (x**2) / (x**2 - 1) * cos(ralpha) * 1.8
        a = x * b * sin(ralpha)
        cell = SC_MCLC(a, b, c, alpha)
    elif lattice_variation == "mclc4":
        b = PI
        x = 1.2
        alpha = 65
        ralpha = alpha * TORADIANS
        c = b * (x**2) / (x**2 - 1) * cos(ralpha)
        a = x * b * sin(ralpha)
        cell = SC_MCLC(a, b, c, alpha)
    elif lattice_variation == "mclc5":
        b = PI
        x = 1.4
        alpha = 53
        ralpha = alpha * TORADIANS
        c = b * (x**2) / (x**2 - 1) * cos(ralpha) * 0.9
        a = x * b * sin(ralpha)
        cell = SC_MCLC(a, b, c, alpha)
    elif lattice_variation in ["tri1a", "tri1", "tri", "tria"]:
        cell = SC_TRI(0.5, 0.5, 0.5, 50.0, 60.0, 80.0)
    elif lattice_variation in ["tri2a", "tri2"]:
        cell = SC_TRI(0.5, 1.0, 1.0, 100.0, 100.0, 140.0)
    elif lattice_variation in ["tri1b", "trib"]:
        cell = SC_TRI(0.5, 1.0, 1.0, 100.0, 110.0, 60.0)
    elif lattice_variation == "tri2b":
        cell = SC_TRI(0.5, 1.0, 2.0, 80.0, 100.0, 40.0)

    return cell


# Deprecated in 0.6.3
# TODO:REMOVE in April of 2025
def sc_get_example_cell(lattice_variation: str = None):
    import warnings

    warnings.warn(
        "wulfric.cell.sc_get_example_cell has been renamed to wulfric.cell.sc_get_example in v0.6.3 release. This function with an old name will be removed in April of 2026.",
        DeprecationWarning,
    )

    return sc_get_example(lattice_variation=lattice_variation)


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from math import cos, sin

import numpy as np

from wulfric.cell._basic_manipulation import get_params
from wulfric.constants._numerical import TORADIANS
from wulfric._exceptions import PotentialBugError

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _get_points_table_69():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
        "M": np.array([1 / 2, 1 / 2, 0.0]),
        "X": np.array([0.0, 1 / 2, 0.0]),
        "X1": np.array([1 / 2, 0.0, 0.0]),
    }


def _get_points_table_70():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "X": np.array([1 / 2, 0.0, 1 / 2]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "W": np.array([1 / 2, 1 / 4, 3 / 4]),
        "W2": np.array([3 / 4, 1 / 4, 1 / 2]),
        "K": np.array([3 / 8, 3 / 8, 3 / 4]),
        "U": np.array([5 / 8, 1 / 4, 5 / 8]),
    }


def _get_points_table_71():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "H": np.array([1 / 2, -1 / 2, 1 / 2]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "N": np.array([0.0, 0.0, 1 / 2]),
    }


def _get_points_table_72():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Z": np.array([0.0, 0.0, 1 / 2]),
        "M": np.array([1 / 2, 1 / 2, 0.0]),
        "A": np.array([1 / 2, 1 / 2, 1 / 2]),
        "R": np.array([0.0, 1 / 2, 1 / 2]),
        "X": np.array([0.0, 1 / 2, 0.0]),
    }


def _get_points_table_73(a, c):
    eta = (1 + (c / a) ** 2) / 4
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "M": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "X": np.array([0.0, 0.0, 1 / 2]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "Z": np.array([eta, eta, -eta]),
        "Z0": np.array([-eta, 1 - eta, eta]),
        "N": np.array([0.0, 1 / 2, 0.0]),
    }


def _get_points_table_74(a, c):
    eta = (1 + (a / c) ** 2) / 4
    zeta = ((a / c) ** 2) / 2
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "M": np.array([1 / 2, 1 / 2, -1 / 2]),
        "X": np.array([0.0, 0.0, 1 / 2]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "N": np.array([0.0, 1 / 2, 0.0]),
        "S0": np.array([-eta, eta, eta]),
        "S": np.array([eta, 1 - eta, -eta]),
        "R": np.array([-zeta, zeta, 1 / 2]),
        "G": np.array([1 / 2, 1 / 2, -zeta]),
    }


def _get_points_table_75():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "X": np.array([1 / 2, 0.0, 0.0]),
        "Z": np.array([0.0, 0.0, 1 / 2]),
        "U": np.array([1 / 2, 0.0, 1 / 2]),
        "Y": np.array([0.0, 1 / 2, 0.0]),
        "S": np.array([1 / 2, 1 / 2, 0.0]),
        "T": np.array([0.0, 1 / 2, 1 / 2]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
    }


def _get_points_table_76(a, b, c):
    zeta = (1 + (a / b) ** 2 - (a / c) ** 2) / 4
    eta = (1 + (a / b) ** 2 + (a / c) ** 2) / 4
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "T": np.array([1.0, 1 / 2, 1 / 2]),
        "Z": np.array([1 / 2, 1 / 2, 0.0]),
        "Y": np.array([1 / 2, 0.0, 1 / 2]),
        "SIGMA0": np.array([0.0, eta, eta]),
        "U0": np.array([1.0, 1 - eta, 1 - eta]),
        "A0": np.array([1 / 2, 1 / 2 + zeta, zeta]),
        "C0": np.array([1 / 2, 1 / 2 - zeta, 1 - zeta]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
    }


def _get_points_table_77(a, b, c):
    zeta = (1 + (c / a) ** 2 - (c / b) ** 2) / 4
    eta = (1 + (c / a) ** 2 + (c / b) ** 2) / 4
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "T": np.array([0.0, 1 / 2, 1 / 2]),
        "Z": np.array([1 / 2, 1 / 2, 1.0]),
        "Y": np.array([1 / 2, 0.0, 1 / 2]),
        "LAMBDA0": np.array([eta, eta, 0.0]),
        "Q0": np.array([1 - eta, 1 - eta, 1.0]),
        "G0": np.array([1 / 2 - zeta, 1 - zeta, 1 / 2]),
        "H0": np.array([1 / 2 + zeta, zeta, 1 / 2]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
    }


def _get_points_table_78(a, b, c):
    eta = (1 + (a / b) ** 2 - (a / c) ** 2) / 4
    delta = (1 + (b / a) ** 2 - (b / c) ** 2) / 4
    phi = (1 + (c / b) ** 2 - (c / a) ** 2) / 4
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "T": np.array([0.0, 1 / 2, 1 / 2]),
        "Z": np.array([1 / 2, 1 / 2, 0.0]),
        "Y": np.array([1 / 2, 0.0, 1 / 2]),
        "A0": np.array([1 / 2, 1 / 2 + eta, eta]),
        "C0": np.array([1 / 2, 1 / 2 - eta, 1 - eta]),
        "B0": np.array([1 / 2 + delta, 1 / 2, delta]),
        "D0": np.array([1 / 2 - delta, 1 / 2, 1 - delta]),
        "G0": np.array([phi, 1 / 2 + phi, 1 / 2]),
        "H0": np.array([1 - phi, 1 / 2 - phi, 1 / 2]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
    }


def _get_points_table_79(a, b, c):
    zeta = (1 + (a / c) ** 2) / 4
    eta = (1 + (b / c) ** 2) / 4
    delta = (b**2 - a**2) / 4 / c**2
    mu = (a**2 + b**2) / 4 / c**2
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "X": np.array([1 / 2, 1 / 2, -1 / 2]),
        "S": np.array([1 / 2, 0.0, 0.0]),
        "R": np.array([0.0, 1 / 2, 0.0]),
        "T": np.array([0.0, 0.0, 1 / 2]),
        "W": np.array([1 / 4, 1 / 4, 1 / 4]),
        "SIGMA0": np.array([-zeta, zeta, zeta]),
        "F2": np.array([zeta, 1 - zeta, -zeta]),
        "Y0": np.array([eta, -eta, eta]),
        "U0": np.array([1 - eta, eta, -eta]),
        "L0": np.array([-mu, mu, 1 / 2 - delta]),
        "M0": np.array([mu, -mu, 1 / 2 + delta]),
        "J0": np.array([1 / 2 - delta, 1 / 2 + delta, -mu]),
    }


def _get_points_table_80(a, b, c):
    zeta = (1 + (b / a) ** 2) / 4
    eta = (1 + (c / a) ** 2) / 4
    delta = (c**2 - b**2) / 4 / a**2
    mu = (b**2 + c**2) / 4 / a**2
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "X": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "S": np.array([1 / 2, 0.0, 0.0]),
        "R": np.array([0.0, 1 / 2, 0.0]),
        "T": np.array([0.0, 0.0, 1 / 2]),
        "W": np.array([1 / 4, 1 / 4, 1 / 4]),
        "Y0": np.array([zeta, -zeta, zeta]),
        "U2": np.array([-zeta, zeta, 1 - zeta]),
        "LAMBDA0": np.array([eta, eta, -eta]),
        "G2": np.array([-eta, 1 - eta, eta]),
        "K": np.array([1 / 2 - delta, -mu, mu]),
        "K2": np.array([1 / 2 + delta, mu, -mu]),
        "K4": np.array([-mu, 1 / 2 - delta, 1 / 2 + delta]),
    }


def _get_points_table_81(a, b, c):
    zeta = (1 + (c / b) ** 2) / 4
    eta = (1 + (a / b) ** 2) / 4
    delta = (a**2 - c**2) / 4 / b**2
    mu = (c**2 + a**2) / 4 / b**2
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "X": np.array([1 / 2, -1 / 2, 1 / 2]),
        "S": np.array([1 / 2, 0.0, 0.0]),
        "R": np.array([0.0, 1 / 2, 0.0]),
        "T": np.array([0.0, 0.0, 1 / 2]),
        "W": np.array([1 / 4, 1 / 4, 1 / 4]),
        "SIGMA0": np.array([-eta, eta, eta]),
        "F0": np.array([eta, -eta, 1 - eta]),
        "LAMBDA0": np.array([zeta, zeta, -zeta]),
        "G0": np.array([1 - zeta, -zeta, zeta]),
        "V0": np.array([mu, 1 / 2 - delta, -mu]),
        "H0": np.array([-mu, 1 / 2 + delta, mu]),
        "H2": np.array([1 / 2 + delta, -mu, 1 / 2 - delta]),
    }


def _get_points_table_82(a, b, c, lattice_type):
    if lattice_type == "oA":
        zeta = (1 + (b / c) ** 2) / 4
    elif lattice_type == "oC":
        zeta = (1 + (a / b) ** 2) / 4
    else:
        raise PotentialBugError(
            error_summary=f'(convention="HPKOT"), table 82. Wrong lattice type, got "{lattice_type}", expected "oA" or "oC".'
        )
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Y": np.array([-1 / 2, 1 / 2, 0.0]),
        "T": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "Z": np.array([0.0, 0.0, 1 / 2]),
        "S": np.array([0.0, 1 / 2, 0.0]),
        "R": np.array([0.0, 1 / 2, 1 / 2]),
        "SIGMA0": np.array([zeta, zeta, 0.0]),
        "C0": np.array([-zeta, 1 - zeta, 0.0]),
        "A0": np.array([zeta, zeta, 1 / 2]),
        "E0": np.array([-zeta, 1 - zeta, 1 / 2]),
    }


def _get_points_table_83(a, b, c, lattice_type):
    if lattice_type == "oA":
        zeta = (1 + (c / b) ** 2) / 4
    elif lattice_type == "oC":
        zeta = (1 + (b / a) ** 2) / 4
    else:
        raise PotentialBugError(
            error_summary=f'(convention="HPKOT"), table 83. Wrong lattice type, got "{lattice_type}", expected "oA" or "oC".'
        )
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Y": np.array([1 / 2, 1 / 2, 0.0]),
        "T": np.array([1 / 2, 1 / 2, 1 / 2]),
        "T2": np.array([1 / 2, 1 / 2, -1 / 2]),
        "Z": np.array([0.0, 0.0, 1 / 2]),
        "Z2": np.array([0.0, 0.0, -1 / 2]),
        "S": np.array([0.0, 1 / 2, 0.0]),
        "R": np.array([0.0, 1 / 2, 1 / 2]),
        "R2": np.array([0.0, 1 / 2, -1 / 2]),
        "DELTA0": np.array([-zeta, zeta, 0.0]),
        "F0": np.array([zeta, 1 - zeta, 0.0]),
        "B0": np.array([-zeta, zeta, 1 / 2]),
        "B2": np.array([-zeta, zeta, -1 / 2]),
        "G0": np.array([zeta, 1 - zeta, 1 / 2]),
        "G2": np.array([zeta, 1 - zeta, -1 / 2]),
    }


def _get_points_table_84():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "A": np.array([0.0, 0.0, 1 / 2]),
        "K": np.array([1 / 3, 1 / 3, 0.0]),
        "H": np.array([1 / 3, 1 / 3, 1 / 2]),
        "H2": np.array([1 / 3, 1 / 3, -1 / 2]),
        "M": np.array([1 / 2, 0.0, 0.0]),
        "L": np.array([1 / 2, 0.0, 1 / 2]),
    }


def _get_points_table_85(a, c):
    delta = ((a / c) ** 2) / 4
    eta = 5 / 6 - 2 * delta
    nu = 1 / 3 + delta

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "T": np.array([1 / 2, 1 / 2, 1 / 2]),
        "L": np.array([1 / 2, 0.0, 0.0]),
        "L2": np.array([0.0, -1 / 2, 0.0]),
        "L4": np.array([0.0, 0.0, -1 / 2]),
        "F": np.array([1 / 2, 0.0, 1 / 2]),
        "F2": np.array([1 / 2, 1 / 2, 0.0]),
        "S0": np.array([nu, -nu, 0.0]),
        "S2": np.array([1 - nu, 0.0, nu]),
        "S4": np.array([nu, 0.0, -nu]),
        "S6": np.array([1 - nu, nu, 0.0]),
        "H0": np.array([1 / 2, -1 + eta, 1 - eta]),
        "H2": np.array([eta, 1 - eta, 1 / 2]),
        "H4": np.array([eta, 1 / 2, 1 - eta]),
        "H6": np.array([1 / 2, 1 - eta, -1 + eta]),
        "M0": np.array([nu, -1 + eta, nu]),
        "M2": np.array([1 - nu, 1 - eta, 1 - nu]),
        "M4": np.array([eta, nu, nu]),
        "M6": np.array([1 - nu, 1 - nu, 1 - eta]),
        "M8": np.array([nu, nu, -1 + eta]),
    }


def _get_points_table_86(a, c):
    zeta = 1 / 6 - ((c / a) ** 2) / 9
    eta = 1 / 2 - 2 * zeta
    nu = 1 / 2 + zeta
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "T": np.array([1 / 2, -1 / 2, 1 / 2]),
        "P0": np.array([eta, -1 + eta, eta]),
        "P2": np.array([eta, eta, eta]),
        "R0": np.array([1 - eta, -eta, -eta]),
        "M": np.array([1 - nu, -nu, 1 - nu]),
        "M2": np.array([nu, -1 + nu, -1 + nu]),
        "L": np.array([1 / 2, 0.0, 0.0]),
        "F": np.array([1 / 2, -1 / 2, 0.0]),
    }


def _get_points_table_87(a, c, beta):
    beta = beta * TORADIANS
    eta = (1 + (a / c) * cos(beta)) / 2 / sin(beta) ** 2
    nu = 1 / 2 + eta * c * cos(beta) / a
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Z": np.array([0.0, 1 / 2, 0.0]),
        "B": np.array([0.0, 0.0, 1 / 2]),
        "B2": np.array([0.0, 0.0, -1 / 2]),
        "Y": np.array([1 / 2, 0.0, 0.0]),
        "Y2": np.array([-1 / 2, 0.0, 0.0]),
        "C": np.array([1 / 2, 1 / 2, 0.0]),
        "C2": np.array([-1 / 2, 1 / 2, 0.0]),
        "D": np.array([0.0, 1 / 2, 1 / 2]),
        "D2": np.array([0.0, 1 / 2, -1 / 2]),
        "A": np.array([-1 / 2, 0.0, 1 / 2]),
        "E": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "H": np.array([-eta, 0.0, 1 - nu]),
        "H2": np.array([-1 + eta, 0.0, nu]),
        "H4": np.array([-eta, 0.0, -nu]),
        "M": np.array([-eta, 1 / 2, 1 - nu]),
        "M2": np.array([-1 + eta, 1 / 2, nu]),
        "M4": np.array([-eta, 1 / 2, -nu]),
    }


def _get_points_table_88(a, b, c, beta):
    beta = beta * TORADIANS
    zeta = (2 + (a / c) * cos(beta)) / 4 / sin(beta) ** 2
    eta = 1 / 2 - 2 * zeta * c * cos(beta) / a
    psi = 3 / 4 - b**2 / 4 / a**2 / sin(beta) ** 2
    phi = psi - (3 / 4 - psi) * a * cos(beta) / c
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Y2": np.array([-1 / 2, 1 / 2, 0.0]),
        "Y4": np.array([1 / 2, -1 / 2, 0.0]),
        "A": np.array([0.0, 0.0, 1 / 2]),
        "M2": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "V": np.array([1 / 2, 0.0, 0.0]),
        "V2": np.array([0.0, 1 / 2, 0.0]),
        "L2": np.array([0.0, 1 / 2, 1 / 2]),
        "C": np.array([1 - psi, 1 - psi, 0.0]),
        "C2": np.array([-1 + psi, psi, 0.0]),
        "C4": np.array([psi, -1 + psi, 0.0]),
        "D": np.array([-1 + phi, phi, 1 / 2]),
        "D2": np.array([1 - phi, 1 - phi, 1 / 2]),
        "E": np.array([-1 + zeta, 1 - zeta, 1 - eta]),
        "E2": np.array([-zeta, zeta, eta]),
        "E4": np.array([zeta, -zeta, 1 - eta]),
    }


def _get_points_table_89(a, b, c, beta):
    beta = beta * TORADIANS
    mu = (1 + (a / b) ** 2) / 4
    delta = -a * c * cos(beta) / 2 / b**2
    zeta = ((a / b) ** 2 + (1 + (a / c) * cos(beta)) / sin(beta) ** 2) / 4
    eta = 1 / 2 - 2 * zeta * c * cos(beta) / a
    phi = 1 + zeta - 2 * mu
    psi = eta - 2 * delta
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Y": np.array([1 / 2, 1 / 2, 0.0]),
        "A": np.array([0.0, 0.0, 1 / 2]),
        "M": np.array([1 / 2, 1 / 2, 1 / 2]),
        "V2": np.array([0.0, 1 / 2, 0.0]),
        "L2": np.array([0.0, 1 / 2, 1 / 2]),
        "F": np.array([-1 + phi, 1 - phi, 1 - psi]),
        "F2": np.array([1 - phi, phi, psi]),
        "F4": np.array([phi, 1 - phi, 1 - psi]),
        "H": np.array([-zeta, zeta, eta]),
        "H2": np.array([zeta, 1 - zeta, 1 - eta]),
        "H4": np.array([zeta, -zeta, 1 - eta]),
        "G": np.array([-mu, mu, delta]),
        "G2": np.array([mu, 1 - mu, -delta]),
        "G4": np.array([mu, -mu, -delta]),
        "G6": np.array([1 - mu, mu, delta]),
    }


def _get_points_table_90(a, b, c, beta):
    beta = beta * TORADIANS
    zeta = ((a / b) ** 2 + (1 + (a / c) * cos(beta)) / sin(beta) ** 2) / 4
    rho = 1 - zeta * (b / a) ** 2
    eta = 1 / 2 - 2 * zeta * c * cos(beta) / a
    mu = eta / 2 + ((a / b) ** 2) / 4 + a * c * cos(beta) / 2 / b**2
    nu = 2 * mu - zeta
    omega = c * (1 - 4 * nu + (a / b) ** 2 * sin(beta) ** 2) / 2 / a / cos(beta)
    delta = -1 / 4 + omega / 2 - zeta * (c / a) * cos(beta)
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Y": np.array([1 / 2, 1 / 2, 0.0]),
        "A": np.array([0.0, 0.0, 1 / 2]),
        "M2": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "V": np.array([1 / 2, 0.0, 0.0]),
        "V2": np.array([0.0, 1 / 2, 0.0]),
        "L2": np.array([0.0, 1 / 2, 1 / 2]),
        "I": np.array([-1 + rho, rho, 1 / 2]),
        "I2": np.array([1 - rho, 1 - rho, 1 / 2]),
        "K": np.array([-nu, nu, omega]),
        "K2": np.array([-1 + nu, 1 - nu, 1 - omega]),
        "K4": np.array([1 - nu, nu, omega]),
        "H": np.array([-zeta, zeta, eta]),
        "H2": np.array([zeta, 1 - zeta, 1 - eta]),
        "H4": np.array([zeta, -zeta, 1 - eta]),
        "N": np.array([-mu, mu, delta]),
        "N2": np.array([mu, 1 - mu, -delta]),
        "N4": np.array([mu, -mu, -delta]),
        "N6": np.array([1 - mu, mu, delta]),
    }


def _get_points_table_91():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Z": np.array([0.0, 0.0, 0.0]),
        "Y": np.array([0, 1 / 2, 0.0]),
        "X": np.array([1 / 2, 0.0, 0.0]),
        "V": np.array([1 / 2, 1 / 2, 0.0]),
        "U": np.array([1 / 2, 0.0, 1 / 2]),
        "T": np.array([0.0, 1 / 2, 1 / 2]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
    }


def _get_points_table_92():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "Z": np.array([0.0, 0.0, 1 / 2]),
        "Y": np.array([0.0, 1 / 2, 0.0]),
        "Y2": np.array([0.0, -1 / 2, 0.0]),
        "X": np.array([1 / 2, 0.0, 0.0]),
        "V2": np.array([1 / 2, -1 / 2, 0.0]),
        "U2": np.array([-1 / 2, 0.0, 1 / 2]),
        "T2": np.array([0, -1 / 2, 1 / 2]),
        "R2": np.array([-1 / 2, -1 / 2, 1 / 2]),
    }


def _hpkot_get_points(conventional_cell, lattice_type, extended_bl_symbol):
    a, b, c, _, beta, _ = get_params(cell=conventional_cell)

    if extended_bl_symbol in ["cP1", "cP2"]:
        return _get_points_table_69()

    if extended_bl_symbol in ["cF1", "cF2"]:
        return _get_points_table_70()

    if extended_bl_symbol == "cI1":
        return _get_points_table_71()

    if extended_bl_symbol == "tP1":
        return _get_points_table_72()

    if extended_bl_symbol == "tI1":
        return _get_points_table_73(a=a, c=c)

    if extended_bl_symbol == "tI2":
        return _get_points_table_74(a=a, c=c)

    if extended_bl_symbol == "oP1":
        return _get_points_table_75()

    if extended_bl_symbol == "oF1":
        return _get_points_table_76(a=a, b=b, c=c)

    if extended_bl_symbol == "oF2":
        return _get_points_table_77(a=a, b=b, c=c)

    if extended_bl_symbol == "oF3":
        return _get_points_table_78(a=a, b=b, c=c)

    if extended_bl_symbol == "oI1":
        return _get_points_table_79(a=a, b=b, c=c)

    if extended_bl_symbol == "oI2":
        return _get_points_table_80(a=a, b=b, c=c)

    if extended_bl_symbol == "oI3":
        return _get_points_table_81(a=a, b=b, c=c)

    if extended_bl_symbol == "oC1":
        return _get_points_table_82(a=a, b=b, c=c, lattice_type=lattice_type)

    if extended_bl_symbol == "oC2":
        return _get_points_table_83(a=a, b=b, c=c, lattice_type=lattice_type)

    if extended_bl_symbol == "oA1":
        return _get_points_table_82(a=a, b=b, c=c, lattice_type=lattice_type)

    if extended_bl_symbol == "oA2":
        return _get_points_table_83(a=a, b=b, c=c, lattice_type=lattice_type)

    if extended_bl_symbol in ["hP1", "hP2"]:
        return _get_points_table_84()

    if extended_bl_symbol == "hR1":
        return _get_points_table_85(a=a, c=c)

    if extended_bl_symbol == "hR2":
        return _get_points_table_86(a=a, c=c)

    if extended_bl_symbol == "mP1":
        return _get_points_table_87(a=a, c=c, beta=beta)

    if extended_bl_symbol == "mC1":
        return _get_points_table_88(a=a, b=b, c=c, beta=beta)

    if extended_bl_symbol == "mC2":
        return _get_points_table_89(a=a, b=b, c=c, beta=beta)

    if extended_bl_symbol == "mC3":
        return _get_points_table_90(a=a, b=b, c=c, beta=beta)

    if extended_bl_symbol == "aP2":
        return _get_points_table_91()

    if extended_bl_symbol == "aP3":
        return _get_points_table_92()

    raise PotentialBugError(
        error_summary=f'(convention="HPKOT"). Unexpected extended Bravais lattice symbol, got "{extended_bl_symbol}".'
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import numpy as np

from wulfric.cell._basic_manipulation import get_reciprocal

try:
    from scipy.spatial import Voronoi

    SCIPY_AVAILABLE = True
except ImportError:
    SCIPY_AVAILABLE = False

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_lattice_points(cell, range=(1, 1, 1), relative=False, flat=True):
    r"""
    Compute lattice points for the given cell.

    Assumes that ``cell`` contains one lattice point.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    range : (3, ) list or tuple of int, default (1, 1, 1)
        How many lattice points to return. All lattice points with relative coordinates
        ``r_1``, ``r_2``, ``r_3``, that fulfil

        * ``-range[0] <= r_1 <= range[0]``
        * ``-range[1] <= r_2 <= range[1]``
        * ``-range[2] <= r_3 <= range[2]``
    relative : bool, default False
        Whether to return relative coordinates.
    flat : bool, default False

    Returns
    -------
    lattice_points : (N1, N2, N3, 3) or (N, 3) :numpy:`ndarray`
        N lattice points. Each element is a vector :math:`v = (v_x, v_y, v_z)`.

        * ``N = N1 * N2 * N3``
        * ``N1 = 2 * range[0] + 1``
        * ``N2 = 2 * range[1] + 1``
        * ``N3 = 2 * range[2] + 1``

        The shape of the returned array is

        * (N, 3) if ``flat=True``
        * (N1, N2, N3, 3) if ``flat=False``
    """

    N1 = 2 * range[0] + 1
    N2 = 2 * range[1] + 1
    N3 = 2 * range[2] + 1

    lp1 = np.arange(-range[0], range[0] + 1)
    lp2 = np.arange(-range[1], range[1] + 1)
    lp3 = np.arange(-range[2], range[2] + 1)

    lattice_points = np.array(np.meshgrid(lp1, lp2, lp3, indexing="ij")).transpose(
        (1, 2, 3, 0)
    )

    if not relative:
        lattice_points = lattice_points @ cell

    if flat:
        lattice_points = np.reshape(lattice_points, (N1 * N2 * N3, 3))

    return lattice_points


def _get_voronoi_cell(cell):
    r"""
    Computes Voronoi edges around (0,0,0) point.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    Returns
    -------
    vertices : (M, 3) :numpy:`ndarray`
        M vertices of the Voronoi cell around (0,0,0) point. Each element is a vector
        :math:`v = (v_x, v_y, v_z)`.
    edges : (N, 2) :numpy:`ndarray`
        N edges of the Voronoi cell around (0,0,0) point. Each elements contains two
        indices of the ``vertices`` forming an edge. Edge ``i`` is between points
        ``vertices[edges[i][0]]`` and ``vertices[edges[i][1]]``.
    """

    n = 10
    index_000 = ((2 * n + 1) ** 3 - 1) // 2

    if not SCIPY_AVAILABLE:
        raise ImportError(
            'SciPy is not available. Please install it with "pip install scipy"'
        )
    voronoi = Voronoi(get_lattice_points(cell, relative=False, range=(n, n, n)))
    edges_index = set()
    # Thanks ASE for the general idea
    # Note that more than -10 0 10 range is required for the lattice points to correctly
    # produce the Voronoi decomposition of the lattice in most cases
    # In general bigger span might be required
    for rv, rp in zip(voronoi.ridge_vertices, voronoi.ridge_points):
        if -1 not in rv and index_000 in rp:
            for j in range(0, len(rv)):
                if (rv[j - 1], rv[j]) not in edges_index and (
                    rv[j],
                    rv[j - 1],
                ) not in edges_index:
                    edges_index.add((rv[j - 1], rv[j]))
    edges_index = np.array(list(edges_index))
    vertices_indices = np.unique(edges_index.flatten())
    vertices_indices_mapping = {
        v_index: index for index, v_index in enumerate(vertices_indices)
    }
    edges = np.zeros((edges_index.shape[0], 2), dtype=int)
    for i in range(edges_index.shape[0]):
        edges[i][0] = vertices_indices_mapping[edges_index[i][0]]
        edges[i][1] = vertices_indices_mapping[edges_index[i][1]]
    return voronoi.vertices[vertices_indices], edges


def get_wigner_seitz_cell(cell):
    r"""
    Computes |Wigner-Seitz|_ cell.

    It assumes that given ``cell`` contains one lattice point.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    Returns
    -------
    vertices : (M, 3) :numpy:`ndarray`
        M vertices of the |Wigner-Seitz|_ cell. Each element is a vector
        :math:`v = (v_x, v_y, v_z)` in absolute (Cartesian) coordinates.
    edges : (N, 2) :numpy:`ndarray`
        N edges of the |Wigner-Seitz|_ cell. Each elements contains two indices of the
        ``vertices`` forming an edge. Edge ``i`` is between points
        ``vertices[edges[i][0]]`` and ``vertices[edges[i][1]]``.
    """

    return _get_voronoi_cell(cell=cell)


def get_brillouin_zone(cell):
    r"""
    Computes Brillouin_zone.

    It assumes that given ``cell`` contains one lattice point.

    Parameters
    ----------
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.

    Returns
    -------
    vertices : (M, 3) :numpy:`ndarray`
        M vertices of the Brillouin_zone. Each element is a vector
        :math:`v = (v_x, v_y, v_z)` in absolute (Cartesian) coordinates.
    edges : (N, 2) :numpy:`ndarray`
        N edges of the Brillouin_zone. Each elements contains two indices of the
        ``vertices`` forming an edge. Edge ``i`` is between points
        ``vertices[edges[i][0]]`` and ``vertices[edges[i][1]]``.
    """

    return _get_voronoi_cell(cell=get_reciprocal(cell=cell))


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from math import cos, pi, sin

import numpy as np
import pytest
from hypothesis import example, given
from hypothesis import strategies as st
from hypothesis.extra.numpy import arrays as harrays

from wulfric._numerical import compare_with_tolerance
from wulfric.constants._numerical import TORADIANS
from wulfric.geometry._geometry import (
    get_angle,
    get_spherical,
    get_volume,
    parallelepiped_check,
)

ANGLE_TOLERANCE = 1e-4
LENGTH_TOLERANCE = 1e-8
MIN_LENGTH = 0.0
MAX_LENGTH = 1e7


################################################################################
#                                     Angle                                    #
################################################################################
@given(
    harrays(float, 3, elements=st.floats(allow_infinity=False, allow_nan=False)),
    harrays(float, 3, elements=st.floats(allow_infinity=False, allow_nan=False)),
)
def test_get_angle(v1, v2):
    if (
        abs(np.linalg.norm(v1)) > np.finfo(float).eps
        and abs(np.linalg.norm(v2)) > np.finfo(float).eps
    ):
        result_degrees = get_angle(v1, v2)
        result_radians = get_angle(v1, v2, radians=True)
        assert 0.0 <= result_degrees <= 180.0
        assert 0.0 <= result_radians <= pi
    else:
        with pytest.raises(ValueError):
            result_degrees = get_angle(v1, v2)


@example(0)
@example(0.0000000001)
@given(st.floats(min_value=-360, max_value=360))
def test_get_angle_values(alpha):
    v1 = np.array([1.0, 0.0, 0.0])
    v2 = np.array([cos(alpha * TORADIANS), sin(alpha * TORADIANS), 0.0])

    if alpha < 0:
        alpha = 360 + alpha

    if alpha > 180:
        alpha = 360 - alpha

    assert abs(get_angle(v1, v2) - alpha) < ANGLE_TOLERANCE


def test_get_angle_raises():
    with pytest.raises(ValueError):
        get_angle([0, 0, 0], [0, 0, 0])
    with pytest.raises(ValueError):
        get_angle([0, 0, 0], [1.0, 0, 0])


################################################################################
#                                    Volume                                    #
################################################################################
@pytest.mark.parametrize(
    "args, result, eps",
    [((4, 4.472, 4.583, 79.03, 64.13, 64.15), 66.3840797, LENGTH_TOLERANCE)],
)
def test_get_volume_example(args, result, eps):
    assert get_volume(*args) - result < eps


# No need to test the vectors - they take the same route as the cell
# if the vectors will move from rows to columns it is not a problem as well
# since the volume of the cell is its determinant
@example(np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
@example(np.array([[-1, 0, 0], [0, 1, 0], [0, 0, 1]]))
@given(
    harrays(
        float, (3, 3), elements=st.floats(min_value=MIN_LENGTH, max_value=MAX_LENGTH)
    )
)
def test_get_volume_with_cell(cell):
    assert get_volume(cell) >= 0


@given(
    st.floats(
        min_value=MIN_LENGTH,
        max_value=MAX_LENGTH,
    ),
    st.floats(
        min_value=MIN_LENGTH,
        max_value=MAX_LENGTH,
    ),
    st.floats(
        min_value=MIN_LENGTH,
        max_value=MAX_LENGTH,
    ),
    st.floats(min_value=0, max_value=360),
    st.floats(min_value=0, max_value=360),
    st.floats(min_value=0, max_value=360),
)
def test_get_volume_parameters(a, b, c, alpha, beta, gamma):
    if parallelepiped_check(a, b, c, alpha, beta, gamma):
        assert get_volume(a, b, c, alpha, beta, gamma) > 0


################################################################################
#                             Parallelepiped check                             #
################################################################################
@given(
    st.floats(min_value=MIN_LENGTH, max_value=MAX_LENGTH),
    st.floats(min_value=MIN_LENGTH, max_value=MAX_LENGTH),
    st.floats(min_value=MIN_LENGTH, max_value=MAX_LENGTH),
    st.floats(min_value=0, max_value=360),
    st.floats(min_value=0, max_value=360),
    st.floats(min_value=0, max_value=360),
)
def test_parallelepiped_check(a, b, c, alpha, beta, gamma):
    assert parallelepiped_check(a, b, c, alpha, beta, gamma) == (
        compare_with_tolerance(a, ">", 0.0, eps=LENGTH_TOLERANCE)
        and compare_with_tolerance(b, ">", 0.0, eps=LENGTH_TOLERANCE)
        and compare_with_tolerance(c, ">", 0.0, eps=LENGTH_TOLERANCE)
        and compare_with_tolerance(alpha, "<", 180.0, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(beta, "<", 180.0, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(gamma, "<", 180.0, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(alpha, ">", 0.0, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(beta, ">", 0.0, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(gamma, ">", 0.0, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(gamma, "<", alpha + beta, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(
            alpha + beta, "<", 360.0 - gamma, eps=ANGLE_TOLERANCE
        )
        and compare_with_tolerance(beta, "<", alpha + gamma, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(
            alpha + gamma, "<", 360.0 - beta, eps=ANGLE_TOLERANCE
        )
        and compare_with_tolerance(alpha, "<", beta + gamma, eps=ANGLE_TOLERANCE)
        and compare_with_tolerance(
            beta + gamma, "<", 360.0 - alpha, eps=ANGLE_TOLERANCE
        )
    )


################################################################################
#                             Spherical coordinates                            #
################################################################################
@given(
    st.floats(min_value=MIN_LENGTH + LENGTH_TOLERANCE, max_value=MAX_LENGTH),
    st.floats(min_value=0.1, max_value=180 - 0.1),
    st.floats(min_value=0.1, max_value=360),
)
def test_get_spherical(r, theta, phi):
    vector = r * np.array(
        [
            np.cos(phi * TORADIANS) * np.sin(theta * TORADIANS),
            np.sin(phi * TORADIANS) * np.sin(theta * TORADIANS),
            np.cos(theta * TORADIANS),
        ]
    )

    c_r, c_theta, c_phi = get_spherical(vector, in_degrees=True)

    assert compare_with_tolerance(theta, "==", c_theta, eps=ANGLE_TOLERANCE)
    assert compare_with_tolerance(phi, "==", c_phi, eps=ANGLE_TOLERANCE)
    assert compare_with_tolerance(r, "==", c_r, eps=LENGTH_TOLERANCE)



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================


from math import cos, sin, tan

import numpy as np

from wulfric.constants._numerical import TORADIANS
from wulfric.constants._sc_convention import SC_BRAVAIS_LATTICE_SHORT_NAMES
from wulfric._exceptions import PotentialBugError
from wulfric.cell import get_params

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def _get_points_table_2():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "M": np.array([1 / 2, 1 / 2, 0]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
        "X": np.array([0, 1 / 2, 0]),
    }


def _get_points_table_3():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "K": np.array([3 / 8, 3 / 8, 3 / 4]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "U": np.array([5 / 8, 1 / 4, 5 / 8]),
        "W": np.array([1 / 2, 1 / 4, 3 / 4]),
        "X": np.array([1 / 2, 0, 1 / 2]),
    }


def _get_points_table_4():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "H": np.array([1 / 2, -1 / 2, 1 / 2]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "N": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_5():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "A": np.array([1 / 2, 1 / 2, 1 / 2]),
        "M": np.array([1 / 2, 1 / 2, 0]),
        "R": np.array([0, 1 / 2, 1 / 2]),
        "X": np.array([0, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_6(a, c):
    eta = (1 + c**2 / a**2) / 4
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "M": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "N": np.array([0, 1 / 2, 0]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "X": np.array([0, 0, 1 / 2]),
        "Z": np.array([eta, eta, -eta]),
        "Z1": np.array([-eta, 1 - eta, eta]),
    }


def _get_points_table_7(a, c):
    eta = (1 + a**2 / c**2) / 4
    zeta = a**2 / (2 * c**2)
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "N": np.array([0, 1 / 2, 0]),
        "P": np.array([1 / 4, 1 / 4, 1 / 4]),
        "SIGMA": np.array([-eta, eta, eta]),
        "SIGMA1": np.array([eta, 1 - eta, -eta]),
        "X": np.array([0, 0, 1 / 2]),
        "Y": np.array([-zeta, zeta, 1 / 2]),
        "Y1": np.array([1 / 2, 1 / 2, -zeta]),
        "Z": np.array([1 / 2, 1 / 2, -1 / 2]),
    }


def _get_points_table_8():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
        "S": np.array([1 / 2, 1 / 2, 0]),
        "T": np.array([0, 1 / 2, 1 / 2]),
        "U": np.array([1 / 2, 0, 1 / 2]),
        "X": np.array([1 / 2, 0, 0]),
        "Y": np.array([0, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_9(a, b, c):
    eta = (1 + a**2 / b**2 + a**2 / c**2) / 4
    zeta = (1 + a**2 / b**2 - a**2 / c**2) / 4

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "A": np.array([1 / 2, 1 / 2 + zeta, zeta]),
        "A1": np.array([1 / 2, 1 / 2 - zeta, 1 - zeta]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "T": np.array([1, 1 / 2, 1 / 2]),
        "X": np.array([0, eta, eta]),
        "X1": np.array([1, 1 - eta, 1 - eta]),
        "Y": np.array([1 / 2, 0, 1 / 2]),
        "Z": np.array([1 / 2, 1 / 2, 0]),
    }


def _get_points_table_10(a, b, c):
    eta = (1 + a**2 / b**2 - a**2 / c**2) / 4
    delta = (1 + b**2 / a**2 - b**2 / c**2) / 4
    phi = (1 + c**2 / b**2 - c**2 / a**2) / 4

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "C": np.array([1 / 2, 1 / 2 - eta, 1 - eta]),
        "C1": np.array([1 / 2, 1 / 2 + eta, eta]),
        "D": np.array([1 / 2 - delta, 1 / 2, 1 - delta]),
        "D1": np.array([1 / 2 + delta, 1 / 2, delta]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "H": np.array([1 - phi, 1 / 2 - phi, 1 / 2]),
        "H1": np.array([phi, 1 / 2 + phi, 1 / 2]),
        "X": np.array([0, 1 / 2, 1 / 2]),
        "Y": np.array([1 / 2, 0, 1 / 2]),
        "Z": np.array([1 / 2, 1 / 2, 0]),
    }


def _get_points_table_11(a, b, c):
    zeta = (1 + a**2 / c**2) / 4
    eta = (1 + b**2 / c**2) / 4
    delta = (b**2 - a**2) / (4 * c**2)
    mu = (a**2 + b**2) / (4 * c**2)

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "L": np.array([-mu, mu, 1 / 2 - delta]),
        "L1": np.array([mu, -mu, 1 / 2 + delta]),
        "L2": np.array([1 / 2 - delta, 1 / 2 + delta, -mu]),
        "R": np.array([0, 1 / 2, 0]),
        "S": np.array([1 / 2, 0, 0]),
        "T": np.array([0, 0, 1 / 2]),
        "W": np.array([1 / 4, 1 / 4, 1 / 4]),
        "X": np.array([-zeta, zeta, zeta]),
        "X1": np.array([zeta, 1 - zeta, -zeta]),
        "Y": np.array([eta, -eta, eta]),
        "Y1": np.array([1 - eta, eta, -eta]),
        "Z": np.array([1 / 2, 1 / 2, -1 / 2]),
    }


def _get_points_table_12(a, b):
    zeta = (1 + a**2 / b**2) / 4

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "A": np.array([zeta, zeta, 1 / 2]),
        "A1": np.array([-zeta, 1 - zeta, 1 / 2]),
        "R": np.array([0, 1 / 2, 1 / 2]),
        "S": np.array([0, 1 / 2, 0]),
        "T": np.array([-1 / 2, 1 / 2, 1 / 2]),
        "X": np.array([zeta, zeta, 0]),
        "X1": np.array([-zeta, 1 - zeta, 0]),
        "Y": np.array([-1 / 2, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_13():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "A": np.array([0, 0, 1 / 2]),
        "H": np.array([1 / 3, 1 / 3, 1 / 2]),
        "K": np.array([1 / 3, 1 / 3, 0]),
        "L": np.array([1 / 2, 0, 1 / 2]),
        "M": np.array([1 / 2, 0, 0]),
    }


def _get_points_table_14(alpha):
    alpha *= TORADIANS

    eta = (1 + 4 * cos(alpha)) / (2 + 4 * cos(alpha))
    nu = 3 / 4 - eta / 2

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "B": np.array([eta, 1 / 2, 1 - eta]),
        "B1": np.array([1 / 2, 1 - eta, eta - 1]),
        "F": np.array([1 / 2, 1 / 2, 0]),
        "L": np.array([1 / 2, 0, 0]),
        "L1": np.array([0, 0, -1 / 2]),
        "P": np.array([eta, nu, nu]),
        "P1": np.array([1 - nu, 1 - nu, 1 - eta]),
        "P2": np.array([nu, nu, eta - 1]),
        "Q": np.array([1 - nu, nu, 0]),
        "X": np.array([nu, 0, -nu]),
        "Z": np.array([1 / 2, 1 / 2, 1 / 2]),
    }


def _get_points_table_15(alpha):
    alpha *= TORADIANS

    eta = 1 / (2 * tan(alpha / 2) ** 2)
    nu = 3 / 4 - eta / 2

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "F": np.array([1 / 2, -1 / 2, 0]),
        "L": np.array([1 / 2, 0, 0]),
        "P": np.array([1 - nu, -nu, 1 - nu]),
        "P1": np.array([nu, nu - 1, nu - 1]),
        "Q": np.array([eta, eta, eta]),
        "Q1": np.array([1 - eta, -eta, -eta]),
        "Z": np.array([1 / 2, -1 / 2, 1 / 2]),
    }


def _get_points_table_16(b, c, alpha):
    alpha *= TORADIANS

    eta = (1 - b * cos(alpha) / c) / (2 * sin(alpha) ** 2)
    nu = 1 / 2 - eta * c * cos(alpha) / b

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "A": np.array([1 / 2, 1 / 2, 0]),
        "C": np.array([0, 1 / 2, 1 / 2]),
        "D": np.array([1 / 2, 0, 1 / 2]),
        "D1": np.array([1 / 2, 0, -1 / 2]),
        "E": np.array([1 / 2, 1 / 2, 1 / 2]),
        "H": np.array([0, eta, 1 - nu]),
        "H1": np.array([0, 1 - eta, nu]),
        "H2": np.array([0, eta, -nu]),
        "M": np.array([1 / 2, eta, 1 - nu]),
        "M1": np.array([1 / 2, 1 - eta, nu]),
        "M2": np.array([1 / 2, eta, -nu]),
        "X": np.array([0, 1 / 2, 0]),
        "Y": np.array([0, 0, 1 / 2]),
        "Y1": np.array([0, 0, -1 / 2]),
        "Z": np.array([1 / 2, 0, 0]),
    }


def _get_points_table_17(a, b, c, alpha):
    alpha *= TORADIANS

    zeta = (2 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
    eta = 1 / 2 + 2 * zeta * c * cos(alpha) / b
    psi = 3 / 4 - a**2 / (4 * b**2 * sin(alpha) ** 2)
    phi = psi + (3 / 4 - psi) * b * cos(alpha) / c

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "N": np.array([1 / 2, 0, 0]),
        "N1": np.array([0, -1 / 2, 0]),
        "F": np.array([1 - zeta, 1 - zeta, 1 - eta]),
        "F1": np.array([zeta, zeta, eta]),
        "F2": np.array([-zeta, -zeta, 1 - eta]),
        "F3": np.array([1 - zeta, -zeta, 1 - eta]),
        "I": np.array([phi, 1 - phi, 1 / 2]),
        "I1": np.array([1 - phi, phi - 1, 1 / 2]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "M": np.array([1 / 2, 0, 1 / 2]),
        "X": np.array([1 - psi, psi - 1, 0]),
        "X1": np.array([psi, 1 - psi, 0]),
        "X2": np.array([psi - 1, -psi, 0]),
        "Y": np.array([1 / 2, 1 / 2, 0]),
        "Y1": np.array([-1 / 2, -1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_18(a, b, c, alpha):
    alpha *= TORADIANS

    mu = (1 + b**2 / a**2) / 4
    delta = b * c * cos(alpha) / (2 * a**2)
    zeta = mu - 1 / 4 + (1 - b * cos(alpha) / c) / (4 * sin(alpha) ** 2)
    eta = 1 / 2 + 2 * zeta * c * cos(alpha) / b
    phi = 1 + zeta - 2 * mu
    psi = eta - 2 * delta

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "F": np.array([1 - phi, 1 - phi, 1 - psi]),
        "F1": np.array([phi, phi - 1, psi]),
        "F2": np.array([1 - phi, -phi, 1 - psi]),
        "H": np.array([zeta, zeta, eta]),
        "H1": np.array([1 - zeta, -zeta, 1 - eta]),
        "H2": np.array([-zeta, -zeta, 1 - eta]),
        "I": np.array([1 / 2, -1 / 2, 1 / 2]),
        "M": np.array([1 / 2, 0, 1 / 2]),
        "N": np.array([1 / 2, 0, 0]),
        "N1": np.array([0, -1 / 2, 0]),
        "X": np.array([1 / 2, -1 / 2, 0]),
        "Y": np.array([mu, mu, delta]),
        "Y1": np.array([1 - mu, -mu, -delta]),
        "Y2": np.array([-mu, -mu, -delta]),
        "Y3": np.array([mu, mu - 1, delta]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_19(a, b, c, alpha):
    alpha *= TORADIANS

    zeta = (b**2 / a**2 + (1 - b * cos(alpha) / c) / sin(alpha) ** 2) / 4
    eta = 1 / 2 + 2 * zeta * c * cos(alpha) / b
    mu = eta / 2 + b**2 / (4 * a**2) - b * c * cos(alpha) / (2 * a**2)
    nu = 2 * mu - zeta
    rho = 1 - zeta * a**2 / b**2
    omega = (4 * nu - 1 - b**2 * sin(alpha) ** 2 / a**2) * c / (2 * b * cos(alpha))
    delta = zeta * c * cos(alpha) / b + omega / 2 - 1 / 4

    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "F": np.array([nu, nu, omega]),
        "F1": np.array([1 - nu, 1 - nu, 1 - omega]),
        "F2": np.array([nu, nu - 1, omega]),
        "H": np.array([zeta, zeta, eta]),
        "H1": np.array([1 - zeta, -zeta, 1 - eta]),
        "H2": np.array([-zeta, -zeta, 1 - eta]),
        "I": np.array([rho, 1 - rho, 1 / 2]),
        "I1": np.array([1 - rho, rho - 1, 1 / 2]),
        "L": np.array([1 / 2, 1 / 2, 1 / 2]),
        "M": np.array([1 / 2, 0, 1 / 2]),
        "N": np.array([1 / 2, 0, 0]),
        "N1": np.array([0, -1 / 2, 0]),
        "X": np.array([1 / 2, -1 / 2, 0]),
        "Y": np.array([mu, mu, delta]),
        "Y1": np.array([1 - mu, -mu, -delta]),
        "Y2": np.array([-mu, -mu, -delta]),
        "Y3": np.array([mu, mu - 1, delta]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_20():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "L": np.array([1 / 2, 1 / 2, 0]),
        "M": np.array([0, 1 / 2, 1 / 2]),
        "N": np.array([1 / 2, 0, 1 / 2]),
        "R": np.array([1 / 2, 1 / 2, 1 / 2]),
        "X": np.array([1 / 2, 0, 0]),
        "Y": np.array([0, 1 / 2, 0]),
        "Z": np.array([0, 0, 1 / 2]),
    }


def _get_points_table_21():
    return {
        "GAMMA": np.array([0.0, 0.0, 0.0]),
        "L": np.array([1 / 2, -1 / 2, 0]),
        "M": np.array([0, 0, 1 / 2]),
        "N": np.array([-1 / 2, -1 / 2, 1 / 2]),
        "R": np.array([0, -1 / 2, 1 / 2]),
        "X": np.array([0, -1 / 2, 0]),
        "Y": np.array([1 / 2, 0, 0]),
        "Z": np.array([-1 / 2, 0, 1 / 2]),
    }


def _sc_get_points(conventional_cell, lattice_type, lattice_variation):
    a, b, c, alpha, _, _ = get_params(cell=conventional_cell)

    if lattice_type not in SC_BRAVAIS_LATTICE_SHORT_NAMES:
        raise PotentialBugError(
            error_summary=f'(convention="SC"). Unexpected lattice type, got "{lattice_type}".'
        )

    lattice_type = SC_BRAVAIS_LATTICE_SHORT_NAMES[lattice_type]

    if lattice_type == "CUB":
        return _get_points_table_2()

    if lattice_type == "FCC":
        return _get_points_table_3()

    if lattice_type == "BCC":
        return _get_points_table_4()

    if lattice_type == "TET":
        return _get_points_table_5()

    if lattice_type == "BCT":
        if lattice_variation == "BCT1":
            return _get_points_table_6(a=a, c=c)
        elif lattice_variation == "BCT2":
            return _get_points_table_7(a=a, c=c)
        else:
            raise PotentialBugError(
                error_summary=f'(convention="SC"), lattice type "BCT". Unexpected lattice variation, got "{lattice_variation}".'
            )

    if lattice_type == "ORC":
        return _get_points_table_8()

    if lattice_type == "ORCF":
        if lattice_variation in ["ORCF1", "ORCF3"]:
            return _get_points_table_9(a=a, b=b, c=c)
        elif lattice_variation == "ORCF2":
            return _get_points_table_10(a=a, b=b, c=c)
        else:
            raise PotentialBugError(
                error_summary=f'(convention="SC"), lattice type "ORCF". Unexpected lattice variation, got "{lattice_variation}".'
            )

    if lattice_type == "ORCI":
        return _get_points_table_11(a=a, b=b, c=c)

    if lattice_type == "ORCC":
        return _get_points_table_12(a=a, b=b)

    if lattice_type == "HEX":
        return _get_points_table_13()

    if lattice_type == "RHL":
        if lattice_variation == "RHL1":
            return _get_points_table_14(alpha=alpha)
        elif lattice_variation == "RHL2":
            return _get_points_table_15(alpha=alpha)
        else:
            raise PotentialBugError(
                error_summary=f'(convention="SC"), lattice type "RHL". Unexpected lattice variation, got "{lattice_variation}".'
            )

    if lattice_type == "MCL":
        return _get_points_table_16(b=b, c=c, alpha=alpha)

    if lattice_type == "MCLC":
        if lattice_variation in ["MCLC1", "MCLC2"]:
            return _get_points_table_17(a=a, b=b, c=c, alpha=alpha)
        elif lattice_variation in ["MCLC3", "MCLC4"]:
            return _get_points_table_18(a=a, b=b, c=c, alpha=alpha)
        elif lattice_variation == "MCLC5":
            return _get_points_table_19(a=a, b=b, c=c, alpha=alpha)
        else:
            raise PotentialBugError(
                error_summary=f'(convention="SC"), lattice type "MCLC". Unexpected lattice variation, got "{lattice_variation}".'
            )

    if lattice_type == "TRI":
        if lattice_variation in ["TRI1a", "TRI2a"]:
            return _get_points_table_20()
        elif lattice_variation in ["TRI1b", "TRI2b"]:
            return _get_points_table_21()
        else:
            raise PotentialBugError(
                error_summary=f'(convention="SC"), lattice type "TRI". Unexpected lattice variation, got "{lattice_variation}".'
            )

    raise PotentialBugError(
        error_summary=f'(convention="SC"). Unexpected lattice type, got "{lattice_type}".'
    )


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
from math import cos, sqrt

import numpy as np

from wulfric._numerical import compare_with_tolerance
from wulfric.constants._numerical import TODEGREES, TORADIANS

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")


def get_volume(*args):
    r"""
    Computes volume.

    Three types of arguments are expected:

    * One argument.
        Matrix ``cell``.
        Volume is computed as:

        .. math::
            V = \boldsymbol{v_1} \cdot (\boldsymbol{v_2} \times \boldsymbol{v_3})
    * Three arguments.
        Vectors ``v1``, ``v2``, ``v3``.
        Volume is computed as:

        .. math::
            V = \boldsymbol{v_1} \cdot (\boldsymbol{v_2} \times \boldsymbol{v_3})
    * Six arguments.
        Parameters ``a``, ``b``, ``c``, ``alpha``, ``beta``, ``gamma``.
        Volume is computed as:

        .. math::
            V = abc\sqrt{1+2\cos(\alpha)\cos(\beta)\cos(\gamma)-\cos^2(\alpha)-\cos^2(\beta)-\cos^2(\gamma)}


    Parameters
    ----------
    v1 : (3,) |array-like|_
        First vector.
    v2 : (3,) |array-like|_
        Second vector.
    v3 : (3,) |array-like|_
        Third vector.
    cell : (3, 3) |array-like|_
        Matrix of a cell, rows are interpreted as vectors.
    a : float, default 1
        Length of the :math:`v_1` vector.
    b : float, default 1
        Length of the :math:`v_2` vector.
    c : float, default 1
        Length of the :math:`v_3` vector.
    alpha : float, default 90
        Angle between vectors :math:`v_2` and :math:`v_3`. In degrees.
    beta : float, default 90
        Angle between vectors :math:`v_1` and :math:`v_3`. In degrees.
    gamma : float, default 90
        Angle between vectors :math:`v_1` and :math:`v_2`. In degrees.

    Returns
    -------
    volume : float
        Volume of corresponding region in space.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.geometry.get_volume([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        1.0
        >>> wulfric.geometry.get_volume([1, 0, 0], [0, 1, 0], [0, 0, 1])
        1.0
        >>> wulfric.geometry.get_volume(1, 1, 1, 90, 90, 90)
        1.0
    """

    if len(args) == 1:
        cell = np.array(args[0])
    elif len(args) == 3:
        cell = np.array(args)
    elif len(args) == 6:
        a, b, c, alpha, beta, gamma = args
        alpha = alpha * TORADIANS
        beta = beta * TORADIANS
        gamma = gamma * TORADIANS
        sq_root = (
            1
            + 2 * cos(alpha) * cos(beta) * cos(gamma)
            - cos(alpha) ** 2
            - cos(beta) ** 2
            - cos(gamma) ** 2
        )
        return a * b * c * sqrt(sq_root)
    else:
        raise ValueError(
            "Unable to identify input. "
            + "Supported: one (3, 3) array-like, or three (3,) array-like, or 6 floats."
        )

    return float(abs(np.linalg.det(cell)))


def get_angle(v1, v2, radians=False):
    r"""
    Computes angle between two vectors.

    .. math::

        \alpha
        =
        \dfrac{(\boldsymbol{v_1} \cdot \boldsymbol{v_2})}
        {\vert\boldsymbol{v_1}\vert\cdot\vert\boldsymbol{v_2}\vert}

    Parameters
    ----------
    v1 : (3,) |array-like|_
        First vector.
    v2 : (3,) |array-like|_
        Second vector.
    radians : bool, default False
        Whether to return value in radians.

    Returns
    -------
    angle: float
        Angle in degrees or radians (see ``radians``).

    Raises
    ------
    ValueError
        If one of the vectors is zero vector (or both). Norm is compared against
        :numpy:`finfo`\ (float).eps.

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.geometry.get_angle([1, 0, 0], [0, 0, 1])
        90.0
        >>> wulfric.geometry.get_angle([1, 0, 0], [1, 0, 0])
        0.0
        >>> round(wulfric.geometry.get_angle([1, 0, 0], [1, 1, 1]), 4)
        54.7356
    """

    # Normalize vectors
    v1_norm = np.linalg.norm(v1)
    v2_norm = np.linalg.norm(v2)
    if abs(v1_norm) <= np.finfo(float).eps or abs(v2_norm) <= np.finfo(float).eps:
        raise ValueError("Angle is ill defined (zero vector).")

    v1 = np.array(v1) / v1_norm
    v2 = np.array(v2) / v2_norm

    alpha = np.arccos(np.clip(np.dot(v1, v2), -1.0, 1.0))
    if radians:
        return float(alpha)

    return float(alpha * TODEGREES)


def parallelepiped_check(
    a,
    b,
    c,
    alpha,
    beta,
    gamma,
    raise_error=False,
    length_tolerance=1e-8,
    angle_tolerance=1e-4,
):
    r"""
    Checks if parallelepiped is valid.

    The conditions are

    * :math:`a > 0`
    * :math:`b > 0`
    * :math:`c > 0`
    * :math:`0 < \alpha < 180`
    * :math:`0 < \beta < 180`
    * :math:`0 < \gamma < 180`
    * :math:`\gamma < \alpha + \beta < 360 - \gamma`
    * :math:`\beta < \alpha + \gamma < 360 - \beta`
    * :math:`\alpha < \beta + \gamma < 360 - \alpha`

    Parameters
    ----------
    a : float
        Length of the :math:`\boldsymbol{v_1}` vector.
    b : float
        Length of the :math:`\boldsymbol{v_2}` vector.
    c : float
        Length of the :math:`\boldsymbol{v_3}` vector.
    alpha : float
        Angle between vectors :math:`\boldsymbol{v_2}` and :math:`\boldsymbol{v_3}`. In degrees.
    beta : float
        Angle between vectors :math:`\boldsymbol{v_1}` and :math:`\boldsymbol{v_3}`. In degrees.
    gamma : float
        Angle between vectors :math:`\boldsymbol{v_1}` and :math:`\boldsymbol{v_2}`. In degrees.
    raise_error : bool, default False
        Whether to raise error if parameters can not form a parallelepiped.
    length_tolerance : float, default :math:`10^{-8}`
        Numerical tolerance for the length variables. Default value is chosen in the
        contexts of condense matter physics, assuming that length is given in Angstroms.
        Please choose appropriate tolerance for your problem.
    angle_tolerance : float, default :math:`10^{-4}`
        Numerical tolerance for the angle variables. Default value is chosen in
        the contexts of condense matter physics, assuming that angles are in degrees.
        Please choose appropriate tolerance for your problem.

    Returns
    -------
    result: bool
        Whether the parameters could from a parallelepiped.

    Raises
    ------
    ValueError
        If parameters can not form a parallelepiped.
        Only raised if ``raise_error`` is ``True`` (it is ``False`` by default).

    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.geometry.parallelepiped_check(1, 1, 1, 90, 90, 90)
        True
        >>> wulfric.geometry.parallelepiped_check(1, 1, 1, 30, 20, 110)
        False
        >>> wulfric.geometry.parallelepiped_check(1, -1, 1, 90, 90, 90)
        False
        >>> wulfric.geometry.parallelepiped_check(1, 0, 1, 90, 90, 90)
        False
        >>> wulfric.geometry.parallelepiped_check(1, 1, 1, 90, 199, 90)
        False

    """

    result = (
        compare_with_tolerance(a, ">", 0.0, eps=length_tolerance)
        and compare_with_tolerance(b, ">", 0.0, eps=length_tolerance)
        and compare_with_tolerance(c, ">", 0.0, eps=length_tolerance)
        and compare_with_tolerance(alpha, "<", 180.0, eps=angle_tolerance)
        and compare_with_tolerance(beta, "<", 180.0, eps=angle_tolerance)
        and compare_with_tolerance(gamma, "<", 180.0, eps=angle_tolerance)
        and compare_with_tolerance(alpha, ">", 0.0, eps=angle_tolerance)
        and compare_with_tolerance(beta, ">", 0.0, eps=angle_tolerance)
        and compare_with_tolerance(gamma, ">", 0.0, eps=angle_tolerance)
        and compare_with_tolerance(gamma, "<", alpha + beta, eps=angle_tolerance)
        and compare_with_tolerance(
            alpha + beta, "<", 360.0 - gamma, eps=angle_tolerance
        )
        and compare_with_tolerance(beta, "<", alpha + gamma, eps=angle_tolerance)
        and compare_with_tolerance(
            alpha + gamma, "<", 360.0 - beta, eps=angle_tolerance
        )
        and compare_with_tolerance(alpha, "<", beta + gamma, eps=angle_tolerance)
        and compare_with_tolerance(
            beta + gamma, "<", 360.0 - alpha, eps=angle_tolerance
        )
    )

    if not result and raise_error:
        message = "Parameters could not form a parallelepiped:\n"
        message += f"a = {a}"
        if not compare_with_tolerance(a, ">", 0.0, eps=length_tolerance):
            message += f" (a <= 0 with numerical tolerance: {length_tolerance})"
        message += "\n"
        message += f"b = {b}"
        if not compare_with_tolerance(b, ">", 0.0, eps=length_tolerance):
            message += f" (b <= 0 with numerical tolerance: {length_tolerance})"
        message += "\n"
        message += f"c = {c}"
        if not compare_with_tolerance(c, ">", 0.0, eps=length_tolerance):
            message += f" (c <= 0 with numerical tolerance: {length_tolerance})"
        message += "\n"
        message += f"alpha = {alpha}\n"
        if not compare_with_tolerance(alpha, "<", 180.0, eps=angle_tolerance):
            message += f"  (alpha >= 180 with numerical tolerance: {angle_tolerance})\n"
        if not compare_with_tolerance(alpha, ">", 0.0, eps=angle_tolerance):
            message += f"  (alpha <= 0 with numerical tolerance: {angle_tolerance})\n"
        message += f"beta = {beta}\n"
        if not compare_with_tolerance(beta, "<", 180.0, eps=angle_tolerance):
            message += f"  (beta >= 180 with numerical tolerance: {angle_tolerance})\n"
        if not compare_with_tolerance(beta, ">", 0.0, eps=angle_tolerance):
            message += f"  (beta <= 0 with numerical tolerance: {angle_tolerance})\n"
        message += f"gamma = {gamma}\n"
        if not compare_with_tolerance(gamma, "<", 180.0, eps=angle_tolerance):
            message += f"  (gamma >= 180 with numerical tolerance: {angle_tolerance})\n"
        if not compare_with_tolerance(gamma, ">", 0.0, eps=angle_tolerance):
            message += f"  (gamma <= 0 with numerical tolerance: {angle_tolerance})\n"
        if not compare_with_tolerance(gamma, "<", alpha + beta, eps=angle_tolerance):
            message += f"Inequality gamma < alpha + beta is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_with_tolerance(
            alpha + beta, "<", 360.0 - gamma, eps=angle_tolerance
        ):
            message += f"Inequality alpha + beta < 360 - gamma is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_with_tolerance(beta, "<", alpha + gamma, eps=angle_tolerance):
            message += f"Inequality beta < alpha + gamma is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_with_tolerance(
            alpha + gamma, "<", 360.0 - beta, eps=angle_tolerance
        ):
            message += f"Inequality alpha + gamma < 360 - beta is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_with_tolerance(alpha, "<", beta + gamma, eps=angle_tolerance):
            message += f"Inequality alpha < beta + gamma is not satisfied with numerical tolerance: {angle_tolerance}\n"
        if not compare_with_tolerance(
            beta + gamma, "<", 360.0 - alpha, eps=angle_tolerance
        ):
            message += f"Inequality beta + gamma < 360 - alpha is not satisfied with numerical tolerance: {angle_tolerance}\n"
        raise ValueError(message)

    return result


def get_spherical(
    vector, in_degrees=True, polar_axis=[0, 0, 1], radial_line_zero=[1, 0, 0]
):
    R"""
    Compute |spherical-coordinates|_ of a vector.

    :math:`(v^x, v^y, v^z) \rightarrow (r, \theta, \phi)`

    Parameters
    ----------
    vector : (3,) |array-like|_
        Vector to be converted.
    in_degrees : bool, default True
        Whether to return angles in degrees or radians. If ``True``, then angles are
        returned in degrees.
    polar_axis : (3,) |array-like|_
        Polar axis (see notes). By default oriented along :math:`+z`.
    radial_line_zero : (3,) |array-like|_
        Zero of the radial line (see notes). By default oriented along :math:`+x`.

    Returns
    -------
    r : float
        Length of the ``vector``.
    polar_angle : float
        Polar angle. Returned in degrees if ``in_degrees = True``, in radians otherwise.
    azimuthal_angle : float
        Azimuthal angle. Returned in degrees if ``in_degrees = True``, in radians
        otherwise.

    Notes
    -----
    ``polar_angle`` is defined as the angle between the polar axis and the given
    ``vector`` with :math:`0 \le \alpha_{polar} \le \pi`.

    Azimuthal angle is defined as the angle of the rotation of the radial line around the
    polar axis. This angle is measured from the ``radial_line_zero`` in accordance to the
    right-hand rule. :math:`0 \le \alpha_{azimuthal} \le 2\pi`.

    Radial line is the projection of the ``vector`` on the plane perpendicular to the
    ``polar_axis``.

    If azimuthal angle is ill-defined, then wulfric returns

    * :math:`0` if polar angle is :math:`0`.
    * :math:`\pi` if polar angle is :math:`\pi`.


    Examples
    --------

    .. doctest::

        >>> import wulfric
        >>> wulfric.geometry.get_spherical([1, 0, 0])
        (1.0, 90.0, 0.0)
        >>> wulfric.geometry.get_spherical([-1, 0, 0])
        (1.0, 90.0, 180.0)
        >>> wulfric.geometry.get_spherical([0, 1, 0])
        (1.0, 90.0, 90.0)
        >>> wulfric.geometry.get_spherical([0, -1, 0])
        (1.0, 90.0, 270.0)
        >>> wulfric.geometry.get_spherical([0, 0, 1])
        (1.0, 0.0, 0.0)
        >>> wulfric.geometry.get_spherical([0, 0, -1])
        (1.0, 180.0, 180.0)
        >>> wulfric.geometry.get_spherical([1, 0, 0], polar_axis=[1, 0, 0])
        (1.0, 0.0, 0.0)

    """
    polar_axis = np.array(polar_axis) / np.linalg.norm(polar_axis)
    radial_line_zero = np.array(radial_line_zero) / np.linalg.norm(radial_line_zero)
    r = float(np.linalg.norm(vector))
    vector = np.array(vector) / r

    if np.allclose(vector, polar_axis):
        polar, azimuthal = 0.0, 0.0
    elif np.allclose(vector, -polar_axis):
        polar, azimuthal = np.pi, np.pi
    else:
        polar = float(np.arccos(np.clip(np.dot(vector, polar_axis), -1, 1)))

        vector = vector - polar_axis * np.dot(vector, polar_axis)
        vector /= np.linalg.norm(vector)

        # This one is more complex, as it is from 0 to 360
        azimuthal = float(np.arccos(np.clip(np.dot(vector, radial_line_zero), -1, 1)))

        if np.linalg.det([polar_axis, radial_line_zero, vector]) >= 0:
            pass
        else:
            azimuthal = 2 * np.pi - azimuthal

    if in_degrees:
        return r, polar * TODEGREES, azimuthal * TODEGREES
    else:
        return r, polar, azimuthal


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
__version__ = "0.6.3"
__doclink__ = "wulfric.org"
__release_date__ = "20 September 2025"


from . import cell, constants, crystal, geometry, io, kpoints
from ._exceptions import *
from ._kpoints_class import *
from ._lepage import *
from ._numerical import *
from ._package_info import *
from ._plotly_engine import *
from ._syntactic_sugar import *
from ._spglib_interface import *



# ================================== LICENSE ===================================
# Wulfric - Cell, Atoms, K-path, visualization.
# Copyright (C) 2023-2025 Andrey Rybakov
#
# e-mail: anry@uv.es, web: adrybakov.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# ================================ END LICENSE =================================
import numpy as np

# Save local scope at this moment
old_dir = set(dir())
old_dir.add("old_dir")

################################################################################
#                      Setyawan and Curtarolo conventions                      #
################################################################################

SC_BRAVAIS_LATTICE_SHORT_NAMES = {
    "cP": "CUB",
    "cF": "FCC",
    "cI": "BCC",
    "tP": "TET",
    "tI": "BCT",
    "oP": "ORC",
    "oF": "ORCF",
    "oI": "ORCI",
    "oA": "ORCC",
    "oC": "ORCC",
    "hP": "HEX",
    "hR": "RHL",
    "mP": "MCL",
    "mC": "MCLC",
    "aP": "TRI",
}

SC_BRAVAIS_LATTICE_LONG_NAMES = {
    "cP": "Cubic",
    "cF": "Face-centered cubic",
    "cI": "Body-centered cubic",
    "tP": "Tetragonal",
    "tI": "Body-centered tetragonal",
    "oP": "Orthorhombic",
    "oF": "Face-centered orthorhombic",
    "oI": "Body-centered orthorhombic",
    "oA": "C-centered orthorhombic",
    "oC": "C-centered orthorhombic",
    "hP": "Hexagonal",
    "hR": "Rhombohedral",
    "mP": "Monoclinic",
    "mC": "C-centered monoclinic",
    "aP": "Triclinic",
}

SC_CONVENTIONAL_TO_PRIMITIVE = {
    "cP": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "cF": np.array(
        [
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
        ]
    ),
    "cI": np.array(
        [
            [-0.5, 0.5, 0.5],
            [0.5, -0.5, 0.5],
            [0.5, 0.5, -0.5],
        ]
    ),
    "tP": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "tI": np.array(
        [
            [-0.5, 0.5, 0.5],
            [0.5, -0.5, 0.5],
            [0.5, 0.5, -0.5],
        ]
    ),
    "oP": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "oF": np.array(
        [
            [0.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 0.5, 0.0],
        ]
    ),
    "oI": np.array(
        [
            [-0.5, 0.5, 0.5],
            [0.5, -0.5, 0.5],
            [0.5, 0.5, -0.5],
        ]
    ),
    "oA": np.array(
        [
            [0.5, 0.5, 0.0],
            [-0.5, 0.5, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "oC": np.array(
        [
            [0.5, 0.5, 0.0],
            [-0.5, 0.5, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "hP": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "hR": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "mP": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "mC": np.array(
        [
            [0.5, -0.5, 0.0],
            [0.5, 0.5, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
    "aP": np.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
        ]
    ),
}

SC_BRAVAIS_LATTICE_VARIATIONS = (
    "CUB",
    "FCC",
    "BCC",
    "TET",
    "BCT1",
    "BCT2",
    "ORC",
    "ORCF1",
    "ORCF2",
    "ORCF3",
    "ORCI",
    "ORCC",
    "HEX",
    "RHL1",
    "RHL2",
    "MCL",
    "MCLC1",
    "MCLC2",
    "MCLC3",
    "MCLC4",
    "MCLC5",
    "TRI1a",
    "TRI2a",
    "TRI1b",
    "TRI2b",
)

SC_DEFAULT_K_PATHS = {
    "CUB": "GAMMA-X-M-GAMMA-R-X|M-R",
    "FCC": "GAMMA-X-W-K-GAMMA-L-U-W-L-K|U-X",
    "BCC": "GAMMA-H-N-GAMMA-P-H|P-N",
    "TET": "GAMMA-X-M-GAMMA-Z-R-A-Z|X-R|M-A",
    "BCT1": "GAMMA-X-M-GAMMA-Z-P-N-Z1-M|X-P",
    "BCT2": "GAMMA-X-Y-SIGMA-GAMMA-Z-SIGMA1-N-P-Y1-Z|X-P",
    "ORC": "GAMMA-X-S-Y-GAMMA-Z-U-R-T-Z|Y-T|U-X|S-R",
    "ORCF1": "GAMMA-Y-T-Z-GAMMA-X-A1-Y|T-X1|X-A-Z|L-GAMMA",
    "ORCF2": "GAMMA-Y-C-D-X-GAMMA-Z-D1-H-C|C1-Z|X-H1|H-Y|L-GAMMA",
    "ORCF3": "GAMMA-Y-T-Z-GAMMA-X-A1-Y|X-A-Z|L-GAMMA",
    "ORCI": "GAMMA-X-L-T-W-R-X1-Z-GAMMA-Y-S-W|L1-Y|Y1-Z",
    "ORCC": "GAMMA-X-S-R-A-Z-GAMMA-Y-X1-A1-T-Y|Z-T",
    "HEX": "GAMMA-M-K-GAMMA-A-L-H-A|L-M|K-H",
    "RHL1": "GAMMA-L-B1|B-Z-GAMMA-X|Q-F-P1-Z|L-P",
    "RHL2": "GAMMA-P-Z-Q-GAMMA-F-P1-Q1-L-Z",
    "MCL": "GAMMA-Y-H-C-E-M1-A-X-H1|M-D-Z|Y-D",
    "MCLC1": "GAMMA-Y-F-L-I|I1-Z-F1|Y-X1|X-GAMMA-N|M-GAMMA",
    "MCLC2": "GAMMA-Y-F-L-I|I1-Z-F1|N-GAMMA-M",
    "MCLC3": "GAMMA-Y-F-H-Z-I-F1|H1-Y1-X-GAMMA-N|M-GAMMA",
    "MCLC4": "GAMMA-Y-F-H-Z-I|H1-Y1-X-GAMMA-N|M-GAMMA",
    "MCLC5": "GAMMA-Y-F-L-I|I1-Z-H-F1|H1-Y1-X-GAMMA-N|M-GAMMA",
    "TRI1a": "X-GAMMA-Y|L-GAMMA-Z|N-GAMMA-M|R-GAMMA",
    "TRI2a": "X-GAMMA-Y|L-GAMMA-Z|N-GAMMA-M|R-GAMMA",
    "TRI1b": "X-GAMMA-Y|L-GAMMA-Z|N-GAMMA-M|R-GAMMA",
    "TRI2b": "X-GAMMA-Y|L-GAMMA-Z|N-GAMMA-M|R-GAMMA",
}


# Populate __all__ with objects defined in this file
__all__ = list(set(dir()) - old_dir)
# Remove all semi-private objects
__all__ = [i for i in __all__ if not i.startswith("_")]
del old_dir
